#!/usr/bin/env python2
# ----------------------------------------------------------------------------
# Copyright (c) 2017 Massachusetts Institute of Technology (MIT)
# All rights reserved.
#
# Distributed under the terms of the BSD 3-clause license.
#
# The full license is in the LICENSE file, distributed with this software.
# ----------------------------------------------------------------------------
"""
:platform: Unix, Mac
:synopsis: This module processes the beacon receiver data in DigitalRF format to
get geophysical parameters and saves the output to digital metadata. Derived from
jitter code from Juha Vierinen.


"""


import os
import sys
import datetime as dt
import argparse
import scipy as sp
import math
import scipy.fftpack as scfft
import scipy.signal as sig
import scipy.constants as s_const
import ephem
import pdb
#plotting
import matplotlib
matplotlib.use('Agg')
import matplotlib.pyplot as plt
import matplotlib.dates as md
from mpl_toolkits.basemap import Basemap
import shutil
# Millstone imports
import digital_rf as drf

TLE_def = ('CASSIOPE','1 39265U 13055A   17015.93102647 +.00001768 +00000-0 +52149-4 0  9992',
                          '2 39265 080.9701 172.4450 0693099 333.2926 023.4045 14.21489964169876')
debug_plot = False

# update_progress() : Displays or updates a console progress bar
## Accepts a float between 0 and 1. Any int will be converted to a float.
## A value under 0 represents a 'halt'.
## A value at 1 or bigger represents 100%
def update_progress(progress):
    barLength = 100 # Modify this to change the length of the progress bar
    status = ""
    if isinstance(progress, int):
        progress = float(progress)
    if not isinstance(progress, float):
        progress = 0
        status = "error: progress var must be float\r\n"
    if progress < 0:
        progress = 0
        status = "Halt...\r\n"
    if progress >= 1:
        progress = 1
        status = "Done...\r\n"
    block = int(round(barLength*progress))
    text = "\rPercent: [{0}] {1}% {2}".format( "#"*block + "-"*(barLength-block), progress*100, status)
    sys.stdout.write(text)
    sys.stdout.flush()

def ephem_doponly(maindir,tleoff=10.):
    """
        This function will output a dictionary that can be used to remove the
        frequency offset.

        Args:
            maindir (:obj:'str'): Directory that holds the digital rf and metadata.
            tleoff (:obj:'float'): Offset of the tle from the actual data.

        Returns:
            outdict (dict[str, obj]): Output data dictionary::

                {
                        't': Time in posix,
                        'dop1': Doppler frequency of 150 MHz channel from TLE ,
                        'dop2': Doppler frequency of 400 MHz channel from TLE ,
                }
    """

    #%% Get Ephem info
    # Assuming this will stay the same
    ut0 = 25567.5
    e2p = 3600.*24#ephem day to utc seconds

    sitepath = os.path.expanduser(os.path.join(maindir, 'metadata/config/site'))
    sitemeta = drf.DigitalMetadataReader(sitepath)
    sdict = sitemeta.read_latest()
    sdict1 = sdict[sdict.keys()[0]]

    infopath = os.path.expanduser(os.path.join(maindir, 'metadata/info'))
    infometa = drf.DigitalMetadataReader(infopath)
    idict = infometa.read_latest()
    idict1 = idict[idict.keys()[0]]

    passpath = os.path.expanduser(os.path.join(maindir, 'metadata/pass/'))
    passmeta = drf.DigitalMetadataReader(passpath)
    pdict = passmeta.read_latest()
    pdict1 = pdict[pdict.keys()[0]]
    rtime = (pdict1['rise_time']-ut0)*e2p
    tsave = pdict.keys()[0]

    Dop_bw = pdict1['doppler_bandwidth']
    t = sp.arange(0,(Dop_bw.shape[0]+1)*10, 10.)+rtime
    t = t.astype(float)

    obsLoc = ephem.Observer()
    obsLoc.lat = sdict1['latitude']
    obsLoc.long = sdict1['longitude']

    satObj = ephem.readtle(idict1['name'], idict1['tle1'][1:-1], idict1['tle2'][1:-1])
    tephem = (t-rtime)*ephem.second + pdict1['rise_time']

    sublat = sp.zeros_like(tephem)
    sublon = sp.zeros_like(tephem)
    for i, itime in enumerate(tephem):
        obsLoc.date = itime
        satObj.compute(obsLoc)
        sublat[i] = sp.rad2deg(satObj.sublat)
        sublon[i] = sp.rad2deg(satObj.sublong)

    # XXX Extend t vector because for the most part the velocities at the edge
    # are not changing much so to avoid having the interpolation extrapolate.
    # If extrapolation used then error messages that the time was off
    t[-1] = t[-1]+600
    t[-2] = t[-2]+500
    t[0] = t[0]-240

    tdop = (t[0:(len(t)-1)]+t[1:len(t)])/2.0
    tdop[0] = tdop[0]-35.0
    # XXX Used this to line up inital TLE times
    tdop = tdop-tleoff

    tephem = (tdop-rtime)*ephem.second + pdict1['rise_time']

    sublat = sp.zeros_like(tephem)
    sublon = sp.zeros_like(tephem)
    for i, itime in enumerate(tephem):
        obsLoc.date = itime
        satObj.compute(obsLoc)
        sublat[i] = sp.rad2deg(satObj.sublat)
        sublon[i] = sp.rad2deg(satObj.sublong)
    return({"t":t, 'tsave':tsave,
            "dop1":sp.interpolate.interp1d(tdop, Dop_bw[:, 0], kind="cubic"),
            "dop2":sp.interpolate.interp1d(tdop, Dop_bw[:, 1], kind="cubic"),
            'sublat':sp.interpolate.interp1d(tdop, sublat, kind="cubic"),
            'sublon':sp.interpolate.interp1d(tdop, sublon, kind="cubic"),
            'site_latitude':float(sdict1['latitude']), 'site_longitude':float(sdict1['longitude'])})

def outlier_removed_fit(m, w = None, n_iter=10, polyord=7):
    """
    Remove outliers using fited data.

    Args:
        m (:obj:`numpy array`): Phase curve.
        n_iter (:obj:'int'): Number of iteration outlier removal
        polyorder (:obj:'int'): Order of polynomial used.

    Returns:
        fit (:obj:'numpy array'): Curve with outliers removed
    """
    if w is None:
        w = sp.ones_like(m)
    W = sp.diag(sp.sqrt(w))
    m2 = sp.copy(m)
    tv = sp.linspace(-1, 1, num=len(m))
    A = sp.zeros([len(m), polyord])
    for j in range(polyord):
        A[:, j] = tv**(float(j))
    A2 = sp.dot(W,A)
    m2w = sp.dot(m2,W)
    fit = None
    for i in range(n_iter):
        xhat = sp.linalg.lstsq(A2, m2w)[0]
        fit = sp.dot(A, xhat)
        # use gradient for central finite differences which keeps order
        resid = sp.gradient(fit - m2)
        std = sp.std(resid)
        bidx = sp.where(sp.absolute(resid) > 2.0*std)[0]
        for bi in bidx:
            A2[bi,:]=0.0
            m2[bi]=0.0
            m2w[bi]=0.0
    if debug_plot:
        plt.plot(m2,label="outlier removed")
        plt.plot(m,label="original")
        plt.plot(fit,label="fit")
        plt.legend()
        plt.ylim([sp.minimum(fit)-std*3.0,sp.maximum(fit)+std*3.0])
        plt.show()
    return(fit)

def open_file(maindir):
    """
        Creates the digital RF reading object.

        Args:
            maindir (:obj:'str'): The directory where the data is located.

        Returns:
            drfObj (obj:"DigitalRFReader"): Digital RF Reader object.
            chandict (obj:"dict"): Dictionary that holds info for the channels.
            start_indx (obj:'long'): Start index in samples.
            end_indx (obj:'long'): End index in samples.
    """
    mainpath = os.path.expanduser(maindir)
    drfObj = drf.DigitalRFReader(mainpath)
    chans = drfObj.get_channels()

    chandict={}

    start_indx, end_indx=[0, sp.inf]
    # Get channel info
    for ichan in chans:
        curdict = {}

        curdict['sind'], curdict['eind'] = drfObj.get_bounds(ichan)
        # determine the read boundrys assuming the sampling is the same.
        start_indx = sp.maximum(curdict['sind'], start_indx)
        end_indx = sp.minimum(curdict['eind'], end_indx)

        dmetadict = drfObj.read_metadata(start_indx, end_indx, ichan)
        dmetakeys = dmetadict.keys()

        curdict['sps'] = dmetadict[dmetakeys[0]]['samples_per_second']
        curdict['fo'] = dmetadict[dmetakeys[0]]['center_frequencies'].ravel()[0]
        chandict[ichan] = curdict

    return (drfObj, chandict, start_indx, end_indx)
def corr_tle_rf(maindir, e=None, window=2**18, n_measure=100,timewin=[0,0],tleoff=0):
    """
        Coorelates the tle derived frequency and the rf frequency. A flag is output
        that specifies if the two frequencies correlate. A time offset between
        the two is also calculated if there is a correlation.

    """
    if e is None:
        e = ephem_doponly(maindir, tleoff)

    bw_search0 = 8e3
    bw_search1 = 20e3
    drfObj, chandict, start_indx, end_indx = open_file(maindir)

    chans = chandict.keys()
    sps = chandict[chans[0]]['sps']
    start_indx = start_indx + timewin[0]*sps
    end_indx = end_indx - timewin[1]*sps
    mid_indx = (end_indx+start_indx)/2
    start_vec = sp.linspace(-150*sps, 150*sps, n_measure)+mid_indx
    tvec = start_vec/sps
    del_f = sps/window
    fvec = sp.arange(-window/2.0, window/2.0)*del_f
    bw_indx0 = int(bw_search0/del_f)
    bw_indx1 = int(bw_search1/del_f)

    fidx0 = sp.arange(-bw_indx0/2, bw_indx0/2, dtype=int) + window/2
    fidx1 = sp.arange(-bw_indx1/2, bw_indx1/2, dtype=int) + window/2

    res0 = sp.zeros([n_measure, window], dtype=float)
    res1 = sp.zeros([n_measure, window], dtype=float)

    freqm0 = sp.zeros([n_measure])
    snr0 = sp.zeros([n_measure])
    freqm1 = sp.zeros([n_measure])
    snr1 = sp.zeros([n_measure])

    wfun = sig.get_window('hann', window)

    subchan = 0
    for i_t, c_st in enumerate(start_vec):
        #pull out data
        z00 = drfObj.read_vector(c_st, window, chans[0])[:, subchan]
        z01 = drfObj.read_vector(c_st+window, window, chans[0])[:, subchan]
        z10 = drfObj.read_vector(c_st, window, chans[1])[:, subchan]
        z11 = drfObj.read_vector(c_st+window, window, chans[1])[:, subchan]

        F0 = scfft.fftshift(scfft.fft(z00*wfun))
        F1 = scfft.fftshift(scfft.fft(z01*wfun))

        res_temp = F0*F1.conj()
        res_noise = sp.median(res_temp.real**2 + res_temp.imag**2)/sp.log(2.)
        res0[i_t, :] = res_temp.real**2 + res_temp.imag**2
        freqm0[i_t] = fvec[fidx0[sp.argmax(res0[i_t, fidx0])]]
        snr0[i_t] = res0[i_t, fidx0].max()/res_noise
        # Cross correlation for second channel and get residual frequency
        F0 = scfft.fftshift(scfft.fft(z10*wfun))
        F1 = scfft.fftshift(scfft.fft(z11*wfun))

        res_temp = F0*F1.conj()
        res_noise = sp.median(res_temp.real**2 + res_temp.imag**2)/sp.log(2.)
        res1[i_t, :] = res_temp.real**2 + res_temp.imag**2
        # find max frequency over restricted sub-band
        freqm1[i_t] = fvec[fidx1[sp.argmax(res1[i_t, fidx1])]]
        snr1[i_t] = res1[i_t, fidx1].max()/res_noise

    thresh = 15
    #outlier removal
    keep0 = 10*sp.log10(snr0) > thresh
    if keep0.sum() == 0:
        return False, 0
    dopfit0 = sp.interpolate.interp1d(tvec[keep0], freqm0[keep0], fill_value='extrapolate')(tvec)
    dfmean0 = sp.mean(dopfit0)
    dfstd0 = sp.std(dopfit0)

    keep1 = 10*sp.log10(snr1) > thresh
    if keep1.sum() == 0:
        return False, 0
    dopfit1 = sp.interpolate.interp1d(tvec[keep1], freqm1[keep1], fill_value='extrapolate')(tvec)
    dfmean1 = sp.mean(dopfit1)
    dfstd1 = sp.std(dopfit1)

    toff = window/sps
    doppler0 = e["dop1"](tvec+toff)
    dmean0 = sp.mean(doppler0)
    dstd0 = sp.std(doppler0)

    doppler1 = e["dop2"](tvec+toff)
    dmean1 = sp.mean(doppler1)
    dstd1 = sp.std(doppler1)

    # cross correlation
    xcor0 = sig.correlate(dopfit0-dfmean0, doppler0-dmean0)/(dfstd0*dstd0*n_measure)
    xcor1 = sig.correlate(dopfit1-dfmean1, doppler1-dmean1)/(dfstd1*dstd1*n_measure)

    if xcor0.max() < .5 or xcor1.max() < .5:
        rfexist = False
        tshift = 0
    else:
        rfexist = True
        corboth = xcor0.max() + xcor1.max()
        lagmean = (xcor0.max()*sp.argmax(xcor0)+xcor1.max()*sp.argmax(xcor1))/corboth
        deltat = tvec[1]-tvec[0]
        tshift = deltat*(n_measure-lagmean)
    return rfexist, tshift, tvec, dopfit0, dopfit1, doppler0, doppler1

def calc_resid(maindir,e,window=2**13,n_measure=500,timewin=[0,0]):
    """
        Calculate the residual difference between the Doppler frequencies from the
        TLEs and measured data.

        Args:
            maindir (:obj:'str'): The directory where the data is located.
            e (:obj:'dict'): The dictionary with information on the output
            window (:obj:'int'): Window length in samples.
            n_window (:obj:'int'): Number of windows integrated.
            bandwidth (:obj:'int'): Number of bins in lowest sub-band to find max frequency.
        Returns:
             outdict (dict[str, obj]): Output data dictionary::

                    {
                            'cspec': Correlated phase residuals,
                            'max_bin': Frequency bin with max return,
                            'doppler_residual': Doppler residual,
                            'tvec': Time vector for each measurment,
                            'fvec': Frequency values array,
                            'res1': Phase residual channel 1,
                            'res0': Phase residual channel 0
                    }
    """
    # number of Hz to search over for max
    bw_search0 = 0.5e3
    bw_search1 = 1e3


    drfObj, chandict, start_indx, end_indx = open_file(maindir)

    chans = chandict.keys()
    sps = chandict[chans[0]]['sps']
    start_indx = start_indx + timewin[0]*sps
    end_indx = end_indx - timewin[1]*sps
    start_vec = sp.linspace(start_indx, end_indx-window*2, n_measure)
    tvec = start_vec/sps
    del_f = sps/window
    fvec = sp.arange(-window/2.0, window/2.0)*del_f

    bw_indx0 = int(bw_search0/del_f)
    bw_indx1 = int(bw_search1/del_f)
    fidx0 = sp.arange(-bw_indx0/2, bw_indx0/2, dtype=int) + window/2
    fidx1 = sp.arange(-bw_indx1/2, bw_indx1/2, dtype=int) + window/2

    res0 = sp.zeros([n_measure, window], dtype=float)
    res1 = sp.zeros([n_measure, window], dtype=float)
    snr0 = sp.zeros(n_measure, dtype=float)
    snr1 = sp.zeros(n_measure, dtype=float)

    freqm0 = sp.zeros([n_measure])
    freqm1 = sp.zeros([n_measure])
    wfun = sig.get_window('hann', window)
    idx = sp.arange(window)
    t_win = idx/sps
    win_s = float(window)/sps
    toff = window/sps
    subchan = 0
    for i_t, c_st in enumerate(start_vec):
        t_cur = tvec[i_t]
        #pull out data
        z00 = drfObj.read_vector(c_st, window, chans[0])[:, subchan]
        z01 = drfObj.read_vector(c_st+window, window, chans[0])[:, subchan]
        z10 = drfObj.read_vector(c_st, window, chans[1])[:, subchan]
        z11 = drfObj.read_vector(c_st+window, window, chans[1])[:, subchan]
        # determine the doppler shift from the sat motion
        # d["e"]["vel2"] is a function derived from the interp1d function
        tphase = sp.float64(t_cur+toff)
        doppler0 = -e["dop1"](tphase)
        doppler1 = -e["dop2"](tphase)
        # Cross correlation for first channel and get residual frequency
        osc00 = wfun*sp.exp(1.0j*2.0*sp.pi*doppler0*t_win)
        osc01 = wfun*sp.exp(1.0j*2.0*sp.pi*doppler0*(t_win + win_s))
        osc10 = wfun*sp.exp(1.0j*2.0*sp.pi*doppler1*t_win)
        osc11 = wfun*sp.exp(1.0j*2.0*sp.pi*doppler1*(t_win + win_s))
        F0 = scfft.fftshift(scfft.fft(z00*osc00.astype(z00.dtype)))
        F1 = scfft.fftshift(scfft.fft(z01*osc01.astype(z01.dtype)))
        res_temp = F0*F1.conj()
        res0[i_t, :] = res_temp.real**2 + res_temp.imag**2
        freqm0[i_t] = fvec[fidx0[sp.argmax(res0[i_t, fidx0])]]
        nc0 = sp.median(res0[i_t,:])/sp.log(2.)
        snr0[i_t] = res0[i_t, fidx0].max()/nc0
        # Cross correlation for second channel and get residual frequency
        F0 = scfft.fftshift(scfft.fft(z10*osc10.astype(z10.dtype)))
        F1 = scfft.fftshift(scfft.fft(z11*osc11.astype(z11.dtype)))
        res_temp = F0*F1.conj()
        res1[i_t, :] = res_temp.real**2 + res_temp.imag**2
        # find max frequency over restricted sub-band
        freqm1[i_t] = fvec[fidx1[sp.argmax(res1[i_t, fidx1])]]
        #normalize residuals
        nc1 = sp.median(res1[i_t,:])/sp.log(2.)
        snr1[i_t] = res1[i_t, fidx1].max()/nc1

        res0[i_t, :] = res0[i_t, :]/nc0
        res1[i_t, :] = res1[i_t, :]/nc1
    tvec[0] = tvec[0] - 100
    tvec[len(tvec)-1] = tvec[len(tvec)-1] + 100
    #outlier removal
    snrmean = .5*snr0+.5*snr1
    dopfit = outlier_removed_fit(.5*(snr0*freqm0*400./150 + snr1*freqm1)/snrmean, snrmean)
    #interpolate residual
    doppler_residual = sp.interpolate.interp1d(tvec, dopfit)
    # correlate residuals together
    rescor = res0*res1.conj()
    #pdb.set_trace()
    cspec = sp.mean(rescor, axis=0)
    return({"cspec":cspec.real, "max_bin":sp.argmax(cspec),
            "doppler_residual":doppler_residual, 'dopfit':dopfit,
            "tvec":tvec, "fvec":fvec, "res1":res1, "res0":res0})

def calc_TEC(maindir, window=4096, incoh_int=100, sfactor=4, offset=0.,timewin=[0,0],snrmin=0.):
    """
    Estimation of phase curve using coherent and incoherent integration.

    Args:
        maindir (:obj:`str`): Path for data.
        window (:obj:'int'): Window length in samples.
        incoh_int (:obj:'int'): Number of incoherent integrations.
        sfactor (:obj:'int'): Overlap factor.
        offset (:obj:'int'): Overlap factor.
        timewin ((:obj:'list'): Overlap factor.)
    Returns:
         outdict (dict[str, obj]): Output data dictionary::

             {
                        "rTEC": Relative TEC in TECU,
                        "rTEC_sig":Relative TEC STD in TECU,
                        "S4": The S4 parameter,
                        "snr0":snr0,
                        "snr1":snr1,
                        "time": Time for each measurement in posix format,
             }
    """

    e = ephem_doponly(maindir, offset)
    resid = calc_resid(maindir, e)
    Nr = int((incoh_int+sfactor-1)*(window/sfactor))

    drfObj, chandict, start_indx, end_indx = open_file(maindir)

    chans = chandict.keys()
    sps = chandict[chans[0]]['sps']
    start_indx = start_indx + timewin[0]*sps
    end_indx = end_indx - timewin[1]*sps
    freq_ratio = chandict[chans[1]]['fo']/chandict[chans[0]]['fo']
    om0,om1 = 2.*s_const.pi*sp.array([chandict[chans[0]]['fo'],chandict[chans[1]]['fo']])
    start_vec = sp.arange(start_indx, end_indx-Nr, Nr, dtype=float)
    tvec = start_vec/sps

    soff = window/sfactor
    toff = soff/sps
    idx = sp.arange(window)
    n_t1 = sp.arange(0, incoh_int)* soff
    IDX, N_t1 = sp.meshgrid(idx, n_t1)
    Msamp = IDX + N_t1
    ls_samp = float(Msamp.flatten()[-1])

    wfun = sig.get_window('hann', window)
    wmat = sp.tile(wfun[sp.newaxis, :], (incoh_int, 1))

    phase_00 = sp.exp(1.0j*0.0)
    phase_10 = sp.exp(1.0j*0.0)

    phase0 = sp.zeros(len(start_vec), dtype=sp.complex64)
    phase1 = sp.zeros(len(start_vec), dtype=sp.complex64)

    phase_cs0 = sp.zeros(len(start_vec),dtype = float)
    phase_cs1 = sp.zeros(len(start_vec),dtype = float)
    snr0 = sp.zeros(len(start_vec))
    snr1 = sp.zeros(len(start_vec))

    std0 = sp.zeros(len(start_vec))
    std1 = sp.zeros(len(start_vec))
    fi = window/2
    subchan = 0
    outspec0 = sp.zeros((len(tvec), window))
    outspec1 = sp.zeros((len(tvec), window))
    print("Start Beacon Processing")
    for i_t, c_st in enumerate(start_vec):

        update_progress(float(i_t)/float(len(start_vec)))
        t_cur = tvec[i_t]

        z00 = drfObj.read_vector(c_st, Nr, chans[0])[Msamp, subchan]
        z01 = drfObj.read_vector(c_st+soff, Nr, chans[0])[Msamp, subchan]
        z10 = drfObj.read_vector(c_st, Nr, chans[1])[Msamp, subchan]
        z11 = drfObj.read_vector(c_st+soff, Nr, chans[1])[Msamp, subchan]

        tphase = sp.float64(t_cur+toff)
        doppler0 = -1.0*(150.0/400.0)*resid["doppler_residual"](t_cur) - e["dop1"](tphase)
        doppler1 = -1.0*resid["doppler_residual"](t_cur) - e["dop2"](tphase)

        osc00 = phase_00*wmat*sp.exp(1.0j*2.0*sp.pi*doppler0*(Msamp/sps))
        osc01 = phase_00*wmat*sp.exp(1.0j*2.0*sp.pi*doppler0*(Msamp/sps+ float(soff)/sps))
        osc10 = phase_10*wmat*sp.exp(1.0j*2.0*sp.pi*doppler1*(Msamp/sps))
        osc11 = phase_10*wmat*sp.exp(1.0j*2.0*sp.pi*doppler1*(Msamp/sps+ float(soff)/sps))


        f00 = scfft.fftshift(scfft.fft(z00*osc00.astype(z00.dtype), axis=-1), axes=-1)
        f01 = scfft.fftshift(scfft.fft(z01*osc01.astype(z01.dtype), axis=-1), axes=-1)
        f00spec = sp.power(f00.real, 2).sum(0)+ sp.power(f00.imag, 2).sum(0)
        outspec0[i_t] = f00spec.real
        f00_cor = f00[:, fi]*sp.conj(f01[:, fi])
        # Use prod to average the phases together.
        phase0[i_t] = sp.cumprod(sp.power(f00_cor,1./float(incoh_int)))[-1]
        phase_cs0[i_t] = sp.cumsum(sp.diff(sp.unwrap(sp.angle(f00[:, fi]))))[-1]

        f10 = scfft.fftshift(scfft.fft(z10*osc10.astype(z10.dtype), axis=-1), axes=-1)
        f11 = scfft.fftshift(scfft.fft(z11*osc11.astype(z11.dtype), axis=-1), axes=-1)
        f10spec = sp.power(f10.real, 2).sum(0)+ sp.power(f10.imag, 2).sum(0)

        f10_cor = f10[:, fi]*sp.conj(f11[:, fi])
        outspec1[i_t] = f10spec.real
        phase1[i_t] = sp.cumprod(sp.power(f10_cor,1./float(incoh_int)))[-1]
        phase_cs1[i_t] = sp.cumsum(sp.diff(sp.unwrap(sp.angle(f10[:, fi]))))[-1]


        std0[i_t] = sp.std(sp.angle(f00_cor))
        std1[i_t] = sp.std(sp.angle(f10_cor))
        snr0[i_t] = f00spec.real[fi]/sp.median(f00spec.real)
        snr1[i_t] = f10spec.real[fi]/sp.median(f10spec.real)

        # Phases for next time through the loop
        phase_00 = phase_00*sp.exp(1.0j*2.0*sp.pi*doppler0*((ls_samp+1.)/sps))

        phase_10 = phase_10*sp.exp(1.0j*2.0*sp.pi*doppler1*((ls_samp+1.)/sps))

    #
    phasecurve = sp.cumsum(sp.angle(phase0)*freq_ratio-sp.angle(phase1))
    phasecurve_amp = phase_cs0*freq_ratio - phase_cs1
    stdcurve = sp.sqrt(sp.cumsum(float(sfactor)*incoh_int*(std0**2.0 + std1**2.0)))

    # SNR windowing, picking values with minimum snr
    snrwin = sp.logical_and(snr0 > snrmin, snr1 > snrmin)
    phasecurve = phasecurve[snrwin]
    phasecurve_amp = phasecurve_amp[snrwin]
    stdcurve = stdcurve[snrwin]
    snr0 = snr0[snrwin]
    snr1 = snr1[snrwin]
    tvec = tvec[snrwin]

    dt = sp.diff(tvec).mean()
    Nside = int(1./dt/2.)
    lvec = sp.arange(-Nside,Nside)
    Lmat, Tmat =sp.meshgrid(lvec, sp.arange(len(tvec)))
    Sampmat = Lmat+Tmat
    Sampclip = sp.clip(Sampmat, 0, len(tvec)-1)
    eps = s_const.e**2/(8.*s_const.pi**2*s_const.m_e*s_const.epsilon_0)
    aconst = s_const.e**2/(2*s_const.m_e*s_const.epsilon_0*s_const.c)
    na = 9.
    nb = 24.
    f0 = 16.668e6

    #cTEC = f0*((na*nb**2)/(na**2-nb**2))*s_const.c/(2.*s_const.pi*eps)
    cTEC = 1e-16*sp.power(om1/om0**2 -1./om1,-1)/aconst
    rTEC = cTEC*phasecurve
    rTEC = rTEC-rTEC.min()
    rTEC_amp = cTEC*phasecurve_amp
    rTEC_amp = rTEC_amp-rTEC_amp.min()
    rTEC_sig = cTEC*stdcurve
    S4 = sp.std(snr0[Sampclip], axis=-1)/sp.median(snr0, axis=-1)


    outdict = {'rTEC':rTEC, 'rTEC_amp':rTEC_amp, 'rTEC_sig':rTEC_sig, 'S4':S4, 'snr0':snr0,
               'snr1':snr1, 'time':tvec, 'resid':resid,'phase':phasecurve,
               'phase_amp':phasecurve_amp, 'phasestd':stdcurve, 'outspec0':outspec0,
               'outspec1':outspec1}
    return outdict
#%% Plotting
def plotsti_vel(maindir, savename='chancomp.png',timewin=[0,0], offset=0, window=512, sfactor=2, incoh_int=10, Nt=512):
    """
        Plot the velocity data over the sti data. This can be used to determie offsets so the data is properly aligned.

        Args:
            maindir (:obj:`str`): Path for data.
            window (:obj:'int'): Window length in samples.
            incoh_int (:obj:'int'): Number of incoherent integrations.
            sfactor (:obj:'int'): Overlap factor.
    """
    # Get the frequency information
    e = ephem_doponly(maindir, offset)


    # always -0th subchannel for beacons
    subchan = 0
    dec_vec = [8, 5]
    flim = [-10, 10]
    mindb = 15.
    Nr = int(sp.prod(dec_vec)*(incoh_int+sfactor-1)*(window/sfactor))

    drfObj, chandict, start_indx, end_indx = open_file(maindir)

    chans = chandict.keys()
    sps = chandict[chans[0]]['sps']
    start_indx = start_indx + timewin[0]*sps
    end_indx = end_indx - timewin[1]*sps
    start_vec = sp.linspace(start_indx, end_indx-Nr, Nt, dtype=float)
    tvec = start_vec/sps
    f0 = e["dop1"](tvec[::4]) * 1e-3
    f1 = e["dop2"](tvec)[::4] * 1e-3

    fvec = sp.arange(-window/2, window/2, dtype=float)*sps/sp.prod(dec_vec)/window

    dates = [dt.datetime.fromtimestamp(ts) for ts in tvec]
    datenums = md.date2num(dates)
    xfmt = md.DateFormatter('%Y-%m-%d %H:%M:%S')

    soff = window/sfactor
    idx = sp.arange(window)
    n_t1 = sp.arange(0, incoh_int)* soff
    IDX, N_t1 = sp.meshgrid(idx, n_t1)
    Msamp = IDX + N_t1

    wfun = sig.get_window('hann', window)
    wmat = sp.tile(wfun[sp.newaxis, :], (incoh_int, 1))

    sti0 = sp.zeros((Nt, window), float)
    sti1 = sp.zeros_like(sti0)
    for i_t, c_st in enumerate(start_vec):

        z0 = drfObj.read_vector(c_st, Nr, chans[0])[:, subchan]
        z1 = drfObj.read_vector(c_st, Nr, chans[1])[:, subchan]
        for idec in dec_vec:
            z0 = sig.decimate(z0, idec)
            z1 = sig.decimate(z1, idec)

        z0 = z0[Msamp]
        z1 = z1[Msamp]

        fft0 = scfft.fftshift(scfft.fft(z0*wmat, axis=-1), axes=-1)
        fft1 = scfft.fftshift(scfft.fft(z1*wmat, axis=-1), axes=-1)

        psd0 = sp.sum(fft0.real**2 + fft0.imag**2, axis=0).real
        psd1 = sp.sum(fft1.real**2 + fft1.imag**2, axis=0).real
        sti0[i_t] = psd0
        sti1[i_t] = psd1


    fig1 = plt.figure(figsize=(12, 12))

    plt.subplot(221)

    mesh = plt.pcolormesh(datenums, fvec*1e-3, sp.transpose(10.*sp.log10(sti0)), vmin=mindb)

    scplot = plt.plot(datenums[::4], f0, 'ko')
    ax = plt.gca()
    ax.xaxis.set_major_formatter(xfmt)
    plt.ylim(flim)
    plt.subplots_adjust(bottom=0.2)
    plt.xticks(rotation=25)
    plt.xlabel("UTC")
    plt.ylabel("Frequency (kHz)")
    plt.title("Power ch0 (dB) %1.2f MHz"%(150.012))
    plt.colorbar(mesh, ax=ax)

    plt.subplot(222)

    mesh = plt.pcolormesh(datenums, fvec*1e-3, sp.transpose(10.*sp.log10(sti0)), vmin=mindb)

    ax = plt.gca()
    ax.xaxis.set_major_formatter(xfmt)
    plt.ylim(flim)
    plt.subplots_adjust(bottom=0.2)
    plt.xticks(rotation=25)
    plt.xlabel("UTC")
    plt.ylabel("Frequency (kHz)")
    plt.title("Power ch0 (dB) %1.2f MHz"%(150.012))
    plt.colorbar(mesh, ax=ax)


    plt.subplot(223)
    mesh = plt.pcolormesh(datenums, fvec*1e-3, sp.transpose(10.*sp.log10(sti1)), vmin=mindb)
    scplot = plt.plot(datenums[::4], f1, 'ko')
    ax = plt.gca()
    ax.xaxis.set_major_formatter(xfmt)
    plt.ylim(flim)
    plt.subplots_adjust(bottom=0.2)
    plt.xticks(rotation=25)
    plt.xlabel("UTC")
    plt.ylabel("Frequency (kHz)")
    plt.title("Power ch1 (dB) %1.2f MHz"%(400.032))
    plt.colorbar(mesh, ax=ax)

    plt.subplot(224)
    mesh = plt.pcolormesh(datenums, fvec*1e-3, sp.transpose(10.*sp.log10(sti1)), vmin=mindb)
    ax = plt.gca()
    ax.xaxis.set_major_formatter(xfmt)
    plt.ylim(flim)
    plt.subplots_adjust(bottom=0.2)
    plt.xticks(rotation=25)
    plt.xlabel("UTC")
    plt.ylabel("Frequency (kHz)")
    plt.title("Power ch1 (dB) %1.2f MHz"%(400.032))
    plt.colorbar(mesh, ax=ax)
    plt.tight_layout()
    print('Saving RF TLE comparison figure: ' + savename)
    fig1.savefig(savename, dpi=300)
    plt.close(fig1)

def plot_resid(d,savename='resfig1.png'):
    """
        Plots the residual frequency after the first wipe using the TLE velocity.
    """
    flim = [-2.e3, 2.e3]
    t = d['tvec']

    dates = [dt.datetime.fromtimestamp(ts) for ts in t]
    datenums = md.date2num(dates)
    xfmt = md.DateFormatter('%Y-%m-%d %H:%M:%S')

    fig1 = plt.figure(figsize=(7, 9))
    doppler_residual = sp.interpolate.interp1d(d['tvec'],d['dopfit'])
    fvec = d["fvec"]
    res0 = d["res0"]
    res1 = d["res1"]
    plt.subplot(211)
    mesh = plt.pcolormesh(datenums, fvec, sp.transpose(10.*sp.log10(res0+1e-12)), vmin=-5, vmax=25)
    plt.plot(datenums, (150.0/400.0)*doppler_residual(t), "r--", label="doppler resid")
    ax = plt.gca()
    ax.xaxis.set_major_formatter(xfmt)
    plt.ylim(flim)
    plt.subplots_adjust(bottom=0.2)
    plt.xticks(rotation=25)
    plt.xlabel("UTC")
    plt.ylabel("Frequency (Hz)")
    plt.title("Power ch0 (dB) %1.2f MHz"%(150.012))
    plt.legend()
    plt.colorbar(mesh, ax=ax)

     # quicklook spectra of residuals spectra along with measured Doppler residual from second channel.
    plt.subplot(212)
    mesh = plt.pcolormesh(datenums, fvec, sp.transpose(10.*sp.log10(res1+1e-12)), vmin=-5, vmax=25)
    plt.plot(datenums, doppler_residual(t), "r--", label="doppler resid")
    ax = plt.gca()
    ax.xaxis.set_major_formatter(xfmt)
    plt.ylim(flim)
    plt.xlabel("UTC")
    plt.ylabel("Frequency (Hz)")
    plt.title("Power ch1 (dB), %1.2f MHz"%(400.032))
    plt.subplots_adjust(bottom=0.2)
    plt.xticks(rotation=25)
    plt.legend()
    plt.colorbar(mesh, ax=ax)

    plt.tight_layout()
    print('Saving residual plots: '+savename)
    plt.savefig(savename, dpi=300)
    plt.close(fig1)

def plot_measurements(outdict, savename='measured.png'):
    """
        Plots the rTEC and S4 measurements.

        Args:
             outdict (dict[str, obj]): Output data dictionary::

                 {
                            "rTEC": Relative TEC in TECU,
                            "rTEC_sig":Relative TEC STD in TECU,
                            "S4": The S4 parameter,
                            "snr0":snr0,
                            "snr1":snr1,
                            "time": Time for each measurement in posix format,
                 }
            savename (obj:'str'): Name of the file that it will be saved to.
    """

    dates = [dt.datetime.fromtimestamp(ts) for ts in outdict['time']]
    datenums = md.date2num(dates)
    xfmt = md.DateFormatter('%Y-%m-%d %H:%M:%S')


    fig1 = plt.figure(figsize=(15, 5))
    plt.subplot(121)
    plt.plot(datenums, outdict['rTEC'], color="black")
    # plt.plot(datenums, outdict['rTEC_amp'], color="black")
    ax = plt.gca()
    ax.grid(True)
    ax.xaxis.set_major_formatter(xfmt)
    plt.xlabel("UTC")
    plt.ylabel("rTEC (TECu)")
    plt.title("Relative TEC and S4 Parameters")
    plt.subplots_adjust(bottom=0.2)
    plt.xticks(rotation=25)
    ax2 = ax.twinx()
    plt.plot(datenums, outdict['S4'], color='red')
    ax2.grid(True)
    ax2.set_ylabel('S4', color='r')
    ax2.tick_params('y', colors='r')

    # SNR from beacon at those times
    plt.subplot(122)
    plt.plot(datenums, 10.0*sp.log10(outdict["snr0"]), label="150 MHz")
    plt.plot(datenums, 10.0*sp.log10(outdict["snr1"]), label="400 MHz")
    plt.legend()
    ax = plt.gca()
    ax.grid(True)
    ax.xaxis.set_major_formatter(xfmt)
    plt.title("SNR for Both Channels")
    plt.xlabel("UTC")
    plt.ylabel("SNR (dB)")
    plt.subplots_adjust(bottom=0.2)
    plt.xticks(rotation=25)

    plt.tight_layout()
    print('Saving measurement plots: ' + savename)
    fig1.savefig(savename, dpi=300)
    plt.close(fig1)



    fig1 = plt.figure(figsize=(15, 5))
    plt.subplot(121)

    plt.plot(datenums, outdict['rTEC_amp'], color="black")
    ax = plt.gca()
    ax.grid(True)
    ax.xaxis.set_major_formatter(xfmt)
    plt.xlabel("UTC")
    plt.ylabel("rTEC (TECu)")
    plt.title("Relative TEC and S4 Parameters")
    plt.subplots_adjust(bottom=0.2)
    plt.xticks(rotation=25)
    ax2 = ax.twinx()
    plt.plot(datenums, outdict['S4'], color='red')
    ax2.grid(True)
    ax2.set_ylabel('S4', color='r')
    ax2.tick_params('y', colors='r')

    # SNR from beacon at those times
    plt.subplot(122)
    plt.plot(datenums, 10.0*sp.log10(outdict["snr0"]), label="150 MHz")
    plt.plot(datenums, 10.0*sp.log10(outdict["snr1"]), label="400 MHz")
    plt.legend()
    ax = plt.gca()
    ax.grid(True)
    ax.xaxis.set_major_formatter(xfmt)
    plt.title("SNR for Both Channels")
    plt.xlabel("UTC")
    plt.ylabel("SNR (dB)")
    plt.subplots_adjust(bottom=0.2)
    plt.xticks(rotation=25)

    plt.tight_layout()
    figpath,name = os.path.split(savename)
    savename = os.path.join(figpath,'amp'+name)
    print('Saving measurement plots: ' + savename)
    fig1.savefig(savename, dpi=300)
    plt.close(fig1)

def plot_map(outdict, e, savename, fig1=None, m=None):
    """
        This function will plot the output data in a scatter plot over a map of the
        satellite path.

        Args:
            outdict (dict[str, obj]): Output dictionary from analyzebeacons.
            e (dict[str, obj]): Output dictionary from ephem_doponly.
            savename(:obj:`str`): Name of the file the image will be saved to.
            fig1(:obj:'matplotlib figure'): Figure.
            m (:obj:'basemap obj'): Basemap object
    """

    t = outdict['time']
    slat = e['site_latitude']
    slon = e['site_longitude']
    if fig1 is None:
        fig1 = plt.figure()
    plt.figure(fig1.number)
    latlim = [math.floor(slat-15.), math.ceil(slat+15.)]
    lonlim = [math.floor(slon-15.), math.ceil(slon+15.)]
    if m is None:
        m = Basemap(lat_0=slat, lon_0=slon, llcrnrlon=lonlim[0], llcrnrlat=latlim[0],
                    urcrnrlon=lonlim[1], urcrnrlat=latlim[1])

    m.drawcoastlines(color="gray")

    m.plot(slon, slat, "rx")
    scat = m.scatter(e["sublon"](t), e["sublat"](t), c=outdict['rTEC'], cmap='viridis',
              vmin=0, vmax=math.ceil(sp.nanmax(outdict['rTEC'])))
    plt.title('Map of TEC Over Satellite Path')
    cb = plt.colorbar(scat,label='rTEC in TECu')
    fig1.savefig(savename, dpi=300)



    #plt.draw()

    scat = m.scatter(e["sublon"](t), e["sublat"](t), c=outdict['rTEC_amp'], cmap='viridis',
              vmin=0, vmax=math.ceil(sp.nanmax(outdict['rTEC_amp'])))
    plt.title('Map of TEC_Amp Over Satellite Path')
    cb.set_clim(vmin=0, vmax=math.ceil(sp.nanmax(outdict['rTEC_amp'])))
    cb.draw_all()

    #plt.tight_layout()
    figpath,name = os.path.split(savename)
    savename = os.path.join(figpath,'amp'+name)
    fig1.savefig(savename, dpi=300)
    plt.close(fig1)

#%% I/O for measurements
def save_output(maindirmeta, outdict, e):
    """
        This function saves the output of the relative TEC measurement processing.

        Args:
            maindir (:obj:`str`): Path for data.
            outdict (dict[str, obj]): Output data dictionary::

                {
                           "rTEC": Relative TEC in TECU,
                           "rTEC_sig":Relative TEC STD in TECU,
                           "S4": The S4 parameter,
                           "snr0":snr0,
                           "snr1":snr1,
                           "time": Time for each measurement in posix format,
                }
        """

    if not os.path.exists(maindirmeta):
        os.mkdir(maindirmeta)

    mdo = drf.DigitalMetadataWriter(
        metadata_dir=maindirmeta,
        subdir_cadence_secs=3600,
        file_cadence_secs=1,
        sample_rate_numerator=1,
        sample_rate_denominator=1,
        file_name='Outputparams',
        )
    # get rid of doppler residual
    if 'doppler_residual' in outdict['resid'].keys():
        del outdict['resid']['doppler_residual']
    mdo.write(e['tsave'], outdict)

def readoutput(maindirmeta):
    """
        This function reads the saved output of the relative TEC measurement processing.
        If the processed data directory doesn't exist it returns None.

        Args:
            maindir (:obj:`str`): Path for RF data.

        Returns:
            outdict (dict[str, obj]): Output data dictionary. If the directory
            doesn't exists it returns None::

                {
                           "rTEC": Relative TEC in TECU,
                           "rTEC_sig":Relative TEC STD in TECU,
                           "S4": The S4 parameter,
                           "snr0":snr0,
                           "snr1":snr1,
                           "time": Time for each measurement in posix format,
                }
        """
    try:
        dmeta = drf.DigitalMetadataReader(maindirmeta)
    except IOError:
        return None
    metadict = dmeta.read_latest()
    outdict = metadict[metadict.keys()[0]]

    return outdict

#%% Run from commandline material
def analyzebeacons(input_args):
    """
        This function will run the analysis code and save the data. Plots will be
        made as well if desired.
    """
    # makes sure theres no trailing / for the path
    mainpath = os.path.expanduser(os.path.dirname(os.path.join(input_args.path, '')))
    maindirmeta = input_args.newdir
    if maindirmeta is None:
        maindirmeta = os.path.join(maindirmeta, 'Processed')
    if not os.path.exists(maindirmeta):
        os.mkdir(maindirmeta)
    figspath = os.path.join(maindirmeta, 'Figures')
    if not os.path.exists(figspath):
        os.mkdir(figspath)
    if input_args.savename is None:
        savename = os.path.join(figspath, 'BeaconPlots.png')
    #rfexist, tleoff = corr_tle_rf(mainpath,timewin=[input_args.begoff, input_args.endoff])
    rfexist = True
    tleoff = input_args.tleoffset
    e = ephem_doponly(mainpath)
    if input_args.justplots or not rfexist:
        print('Analysis will not be run, only plots will be made.')
        plotsti_vel(mainpath, savename=os.path.join(figspath, 'chancomp.png'),
                    timewin=[input_args.begoff, input_args.endoff],
                    offset=tleoff)
        outdict = readoutput(maindirmeta)
        if outdict is None:
            print('No ouptut data exists')
        else:
            plot_resid(outdict['resid'], os.path.join(figspath, 'resid.png'))
            plot_measurements(outdict, savename)
            plot_map(outdict, e, os.path.join(figspath,'mapdata.png'))
    else:

        outdict = calc_TEC(mainpath, window=input_args.window,
                           incoh_int=input_args.incoh, sfactor=input_args.overlap,
                           offset=tleoff, timewin=[input_args.begoff, input_args.endoff],
                           snrmin=input_args.minsnr)
        print("Saving everything to digital metadata.")
        outdict['window'] = input_args.window
        outdict['incoherent_integrations'] = input_args.incoh
        outdict['Overlap'] = input_args.overlap
        outdict['Time_Offset'] = tleoff
        outdict['Beginning_Offset'] = input_args.begoff
        outdict['Ending_Offset'] = input_args.endoff
        outdict['Min_SNR'] = input_args.minsnr
        if os.path.exists(maindirmeta):
            shutil.rmtree(maindirmeta)
        save_output(maindirmeta, outdict, e)

        if input_args.drawplots:
            print("Plotting data.")
            if not os.path.exists(figspath):
                os.mkdir(figspath)
            plotsti_vel(mainpath, savename=os.path.join(figspath, 'chancomp.png'),
                        timewin=[input_args.begoff, input_args.endoff],
                        offset=tleoff)
            plot_resid(outdict['resid'], os.path.join(figspath, 'resid.png'))
            plot_measurements(outdict, savename)
            plot_map(outdict, e, os.path.join(figspath, 'mapdata.png'))

def parse_command_line(str_input=None):
    """
        This will parse through the command line arguments
    """
    # if str_input is None:
    parser = argparse.ArgumentParser()
    # else:
    #     parser = argparse.ArgumentParser(str_input)
    parser.add_argument("-v", "--verbose", action="store_true",
                        dest="verbose", default=False,
                        help="prints debug output and additional detail.")
    parser.add_argument("-c", "--config", dest="config", default='beacons.ini',
                        help="Use configuration file <config>.")
    parser.add_argument('-p', "--path", dest='path',
                        default=None, help='Path to the Digital RF files and meta data.')
    parser.add_argument('-d', "--drawplots", default=False, dest='drawplots', action="store_true",
                        help="Bool to determine if plots will be made and saved.")
    parser.add_argument('-s', "--savename", dest='savename', default=None,
                        help='Name of plot file.')
    parser.add_argument('-w', "--window", dest='window', default=4096, type=int,
                        help='Length of window in samples for FFT in calculations.')
    parser.add_argument('-i', "--incoh", dest='incoh', default=100, type=int,
                        help='Number of incoherent integrations in calculations.')
    parser.add_argument('-o', "--overlap", dest='overlap', default=4, type=int,
                        help='Overlap for each of the FFTs.')
    parser.add_argument('-t', "--tleoffset", dest='tleoffset', default=0., type=float,
                        help="Offset of the TLE time from the actual pass.")
    parser.add_argument('-b', "--begoff", dest='begoff', default=0., type=float,
                        help="Number of seconds to jump ahead before measuring.")
    parser.add_argument('-e', "--endoff", dest='endoff', default=0., type=float,
                        help="Number of seconds to jump ahead before measuring.")
    parser.add_argument('-m', "--minsnr", dest='minsnr', default=0., type=float,
                        help="Minimum SNR for for phase curve measurement")
    parser.add_argument('-j', "--justplots", action="store_true",
                        dest="justplots", default=False,
                        help="Makes plots for input, residuals, and final measurements if avalible.")
    parser.add_argument('-n', "--newdir", dest="newdir", default=None,
                        help='Directory that measured data will be saved.')

    if str_input is None:
        return parser.parse_args()
    else:
        return parser.parse_args(str_input)



if __name__ == '__main__':
    """
        Main way run from command line
    """
    args_commd = parse_command_line()

    if args_commd.path is None:
        print "Please provide an input source with the -p option!"
        sys.exit(1)

    analyzebeacons(args_commd)



# ----------------------------------------------------------------------------
# Copyright (c) 2017 Massachusetts Institute of Technology (MIT)
# All rights reserved.
#
# Distributed under the terms of the BSD 3-clause license.
#
# The full license is in the LICENSE file, distributed with this software.
# ----------------------------------------------------------------------------
"""regression_test_digital_rf_hdf5.py is a script to regression test the digital_rf_hdf5 module

Exits with 0 if successful test, non-zero if failure

$Id$
"""
# standard python imports
import os
import sys
import time
import collections

# third party imports
import numpy

# Millstone imports
import digital_rf

# base data
base_data = []
for i in range(100):
    if i % 7 == 0:
        base_data.append([2, 3])
    else:
        base_data.append([4, 6])

# simulate 10 samples of data followed by 10 sample gap
global_sample_arr = numpy.array(range(10), dtype=numpy.uint64) * 20
block_sample_arr = numpy.array(range(10), dtype=numpy.uint64) * 10

# constants
sample_rate_numerator = int(100)
sample_rate_denominator = 1
sample_rate = numpy.longdouble(sample_rate_numerator) / sample_rate_denominator
subdir_cadence_secs = 10
file_cadence_millisecs = 400
# start 2014-03-09 12:30:30 plus one sample
start_global_index = int(numpy.uint64(1394368230 * sample_rate)) + 1

# test get unix time
dt, picoseconds = digital_rf.get_unix_time(
    int(start_global_index), sample_rate_numerator, sample_rate_denominator)
print('For start_global_index=%i and sample_rate=%f, dt is %s and picoseconds is %i' % (start_global_index, sample_rate,
                                                                                        dt, picoseconds))

# set up top level directory
os.system("rm -rf /tmp/hdf5 ; mkdir /tmp/hdf5")
os.system("rm -rf /tmp/hdf52 ; mkdir /tmp/hdf52")

print("Test 0 - simple single write to multiple files, no compress, no checksum - channel 0")
os.system("rm -rf /tmp/hdf5/junk0 ; mkdir /tmp/hdf5/junk0")
data_object = digital_rf.DigitalRFWriter("/tmp/hdf5/junk0", 'i4', subdir_cadence_secs, file_cadence_millisecs, start_global_index,
                                         sample_rate_numerator, sample_rate_denominator, "FAKE_UUID_0", 0, False, True)
data = numpy.array(base_data, numpy.int32)
result = data_object.rf_write(data)
data_object.close()
print("done test 0")

print("Test 1 - use complex 1 byte ints, no compress, no checksum - channel 1")
os.system("rm -rf /tmp/hdf5/junk1 ; mkdir /tmp/hdf5/junk1")
data_object = digital_rf.DigitalRFWriter("/tmp/hdf5/junk1", 'i1', subdir_cadence_secs, file_cadence_millisecs, start_global_index,
                                         sample_rate_numerator, sample_rate_denominator, "FAKE_UUID_1", 0, False, True)
data = numpy.array(base_data, numpy.int8)
data_object.rf_write(data)
data_object.close()
print(result)
print("done test 1")

print("Test 1.1 - use single 1 byte ints with data gap, no compress, no checksum - channel 1.1")
os.system("rm -rf /tmp/hdf5/junk1.1 ; mkdir /tmp/hdf5/junk1.1")
data_object = digital_rf.DigitalRFWriter("/tmp/hdf5/junk1.1", 'i1', subdir_cadence_secs, file_cadence_millisecs, start_global_index,
                                         sample_rate_numerator, sample_rate_denominator, "FAKE_UUID_1.1", 0, False, False)
data = numpy.array(base_data[:], numpy.int8)
data = data[:, 0]
data_object.rf_write(data)
data_object.rf_write(data, 120)
data_object.close()
print("done test 1.1")

print("Test 1.2 - use single 1 byte ints with no gaps across multiple subdirectories and top_level dirs, no compress, no checksum - channel 1.1")
os.system("rm -rf /tmp/hdf5/junk1.2 ; mkdir /tmp/hdf5/junk1.2")
data_object = digital_rf.DigitalRFWriter("/tmp/hdf5/junk1.2", 'i1', subdir_cadence_secs, file_cadence_millisecs, start_global_index,
                                         sample_rate_numerator, sample_rate_denominator, "FAKE_UUID_1.2", 0, False, False)
data = numpy.array(base_data[:], numpy.int8)
data = data[:, 0]
for i in range(100):
    data_object.rf_write(data)
data_object.close()
print('now use top level dir /tmp/hdf52')
os.system("rm -rf /tmp/hdf52/junk1.2 ; mkdir /tmp/hdf52/junk1.2")
data_object = digital_rf.DigitalRFWriter("/tmp/hdf52/junk1.2", 'i1', subdir_cadence_secs, file_cadence_millisecs, start_global_index + 100 * 100,
                                         sample_rate_numerator, sample_rate_denominator, "FAKE_UUID_1.2", 0, False, False)

data = numpy.array(base_data[:], numpy.int8)
data = data[:, 0]
for i in range(100):
    data_object.rf_write(data)
data_object.close()
print("done test 1.2")

print("Test 2 - use 2 byte ints with data gap, level 1 compress, but no checksum - channel 2")
os.system("rm -rf /tmp/hdf5/junk2 ; mkdir /tmp/hdf5/junk2")
data_object = digital_rf.DigitalRFWriter("/tmp/hdf5/junk2", 'i2', subdir_cadence_secs, file_cadence_millisecs, start_global_index,
                                         sample_rate_numerator, sample_rate_denominator, "FAKE_UUID_2", 1, False, True, is_continuous=False)
data = numpy.array(base_data, numpy.int16)
data_object.rf_write(data)
data_object.rf_write(data, 120)
data_object.close()
print("done test 2")

print("Test 4.1 - use single 8 byte ints with 10 on/10 missing blocks, both compress (level 6) and checksum - channel 4.1")
os.system("rm -rf /tmp/hdf5/junk4.1 ; mkdir /tmp/hdf5/junk4.1")
data_object = digital_rf.DigitalRFWriter("/tmp/hdf5/junk4.1", 'i8', subdir_cadence_secs, file_cadence_millisecs, start_global_index,
                                         sample_rate_numerator, sample_rate_denominator, "FAKE_UUID_2", 1, False, True, is_continuous=False)
data = numpy.array(base_data, numpy.int64)
for i in range(100):
    data_object.rf_write_blocks(data, global_sample_arr, block_sample_arr)
    global_sample_arr += 205
data_object.close()

print('Now write more of this same test 4.1 to top level directory /tmp/hdf52')
os.system("rm -rf /tmp/hdf52/junk4.1 ; mkdir /tmp/hdf52/junk4.1")
data_object = digital_rf.DigitalRFWriter("/tmp/hdf52/junk4.1", 'i8', subdir_cadence_secs, file_cadence_millisecs, start_global_index,
                                         sample_rate_numerator, sample_rate_denominator, "FAKE_UUID_2", 1, False, True, is_continuous=False)
data = numpy.array(base_data, numpy.int64)
for i in range(100):
    data_object.rf_write_blocks(data, global_sample_arr, block_sample_arr)
    global_sample_arr += 205
data_object.close()
print("done test 4.1")

# sleep for 4 seconds to make sure system knows all files closed
time.sleep(4)

# read
t = time.time()
testReadObj = digital_rf.DigitalRFReader(['/tmp/hdf5', '/tmp/hdf52'])
print('init took %f' % (time.time() - t))
channels = testReadObj.get_channels()
print('existing channels are <%s>' % (str(channels)))

chan_name = 'junk4.1'
print('\nworking on channel %s' % (chan_name))
start_index, end_index = testReadObj.get_bounds(chan_name)
print('Bounds are %i to %i' % (start_index, end_index))

# test one of get_continuous_blocks - end points in gaps
start_sample = 139436843434L
end_sample = 139436843538L
print('\ncalling get_continuous_blocks between %i and %i (edges in data gaps)' %
      (start_sample, end_sample))
cont_data_arr = testReadObj.get_continuous_blocks(
    start_sample, end_sample, chan_name)
corr_result = collections.OrderedDict([(139436843436L, 10L),
                                       (139436843456L, 10L),
                                       (139436843476L, 10L),
                                       (139436843501L, 10L),
                                       (139436843521L, 10L)])
if not numpy.array_equal(cont_data_arr, corr_result):
    raise ValueError, 'cont_data_arr = <%s>, but corr_result = <%s>' % (
        str(cont_data_arr), str(corr_result))
print('Got correct result %s' % (str(cont_data_arr)))

# test two of get_continuous_blocks - end points in data
start_sample = 139436843438L
end_sample = 139436843528L
print('\ncalling get_continuous_blocks between %i and %i (edges in continuous data)' %
      (start_sample, end_sample))
cont_data_arr = testReadObj.get_continuous_blocks(
    start_sample, end_sample, chan_name)
corr_result = collections.OrderedDict([(139436843438, 8),
                                       (139436843456, 10),
                                       (139436843476, 10),
                                       (139436843501, 10),
                                       (139436843521, 8)])
if not numpy.array_equal(cont_data_arr, corr_result):
    raise ValueError, 'cont_data_arr = <%s>, but corr_result = <%s>' % (
        str(cont_data_arr), str(corr_result))
print('Got correct result %s' % (str(cont_data_arr)))

# normal read
print('\ntest of read of continuous data')
first_key = cont_data_arr.keys()[0]
print((first_key, cont_data_arr[first_key]))
read_result = testReadObj.read_vector(
    first_key, cont_data_arr[first_key], chan_name)
corr_result = numpy.array([[4.0 + 6.0j],
                           [4.0 + 6.0j],
                           [4.0 + 6.0j],
                           [4.0 + 6.0j],
                           [4.0 + 6.0j],
                           [2.0 + 3.0j],
                           [4.0 + 6.0j],
                           [4.0 + 6.0j]])
if not numpy.array_equal(read_result, corr_result):
    raise ValueError, 'read_result = <%s>, but corr_result = <%s>' % (
        str(read_result), str(corr_result))
print('Got correct result %s' % (str(read_result)))

print('\nnow try reading across a data gap, which should raise an exception')
try:
    # read too far
    result = testReadObj.read_vector(first_key, cont_data_arr[
                                     first_key] + 1, chan_name)
    raise ValueError, 'whoops - no error when one expected!!!!!'
except IOError:
    print(sys.exc_info()[1])
    print('got expected error')


# read one less than block - should be okay
print('\ntest of reading subset of a continuous block')
read_result = testReadObj.read_vector(
    first_key, cont_data_arr[first_key] - 1, chan_name)
if not numpy.array_equal(read_result, corr_result[:-1]):
    raise ValueError, 'read_result = <%s>, but corr_result = <%s>' % (
        str(read_result), str(corr_result[:-1]))
print('Got correct result %s' % (str(read_result)))

chan_name = 'junk1.2'
print('\n now working on channel %s - large continuous data, with padding because continuous but not at file boundaries' % (chan_name))
start_index, end_index = testReadObj.get_bounds(chan_name)
corr_result = (139436823000L, 139436843039L)
if start_index != corr_result[0] or end_index != corr_result[1]:
    raise ValueError, 'got wrong bounds: %s versus %s' % (
        str((start_index, end_index)), str(corr_result))
print('got correct bounds %s' % (str((start_index, end_index))))

print('\ntest of get_continuous_blocks for continuous data')
result = testReadObj.get_continuous_blocks(start_index, end_index, chan_name)
key = result.keys()[0]
if not len(result.keys()) == 1 or key != corr_result[0] or result[key] != (corr_result[1] - corr_result[0]) + 1:
    raise ValueError, 'got wrong continuous blocks: %s versus %s' % (
        str(result), str(corr_result))
print('get correct continuous blocks %s' % (str(result)))

print('\ntest of read_vector for all continuous data')
t = time.time()
result = testReadObj.read_vector(
    start_index, end_index - start_index, chan_name)
if len(result) != corr_result[1] - corr_result[0]:
    raise ValueError, 'got length %i, expected %i' % (
        len(result), corr_result[1] - corr_result[0])
print('got expected len %i' % (len(result)))
print('read took %f' % (time.time() - t))


print('regression test of digital_rf_hdf5 SUCCESS')



# ----------------------------------------------------------------------------
# Copyright (c) 2017 Massachusetts Institute of Technology (MIT)
# All rights reserved.
#
# Distributed under the terms of the BSD 3-clause license.
#
# The full license is in the LICENSE file, distributed with this software.
# ----------------------------------------------------------------------------
"""Module for watching a directory and deleting the oldest Digital RF files."""

import datetime
import os
import re
import sys
import threading
import time
import traceback
from collections import OrderedDict, defaultdict, deque, namedtuple

from . import list_drf, util
from .list_drf import ilsdrf
from .watchdog_drf import DigitalRFEventHandler, DirWatcher

__all__ = (
    'DigitalRFRingbufferHandler', 'DigitalRFRingbuffer',
)


class DigitalRFRingbufferHandlerBase(DigitalRFEventHandler):
    """Base event handler for implementing a ringbuffer of Digital RF files.

    This handler tracks files but does nothing to expire them. At least one
    expirer mixin must be used with this class in order to create a complete
    ringbuffer.

    """

    FileRecord = namedtuple('FileRecord', ('key', 'size', 'path', 'group'))

    def __init__(
        self, verbose=False, dryrun=False, starttime=None, endtime=None,
        include_drf=True, include_dmd=True,
    ):
        """Create a ringbuffer handler.

        Other Parameters
        ----------------

        starttime : datetime.datetime
            Data covering this time or after will be included. This has no
            effect on property files.

        endtime : datetime.datetime
            Data covering this time or earlier will be included. This has no
            effect on property files.

        include_drf : bool
            If True, include Digital RF files.

        include_dmd : bool
            If True, include Digital Metadata files.

        """
        self.verbose = verbose
        self.dryrun = dryrun
        # separately track file groups (ch path, name) with different queues
        self.queues = defaultdict(deque)
        self.records = {}
        # acquire the record lock to modify the queue or record dicts
        self._record_lock = threading.RLock()
        super(DigitalRFRingbufferHandlerBase, self).__init__(
            starttime=starttime, endtime=endtime,
            include_drf=include_drf, include_dmd=include_dmd,
            include_drf_properties=False, include_dmd_properties=False,
        )

    def status(self):
        """Return status string about state of the ringbuffer."""
        nfiles = sum(len(q) for q in self.queues.values())
        return '{0} files'.format(nfiles)

    def _get_file_record(self, path):
        """Return self.FileRecord tuple for file at path."""
        # get time key (seconds) from file path
        key = None
        for r in self.regexes:
            m = r.match(path)
            try:
                secs = int(m.group('secs'))
            except (AttributeError, IndexError, TypeError):
                # no match, or regex matched but there is no 'secs' in regex
                continue
            else:
                try:
                    frac = int(m.group('frac'))
                except (IndexError, TypeError):
                    frac = 0
                # key is time in milliseconds
                key = secs*1000 + frac
                break
        if key is None:
            return

        # ringbuffer by file groups, which are a channel path and name
        group = (m.group('chpath'), m.group('name'))

        try:
            stat = os.stat(path)
        except OSError:
            if self.verbose:
                traceback.print_exc()
            return
        else:
            size = stat.st_size

        return self.FileRecord(key=key, size=size, path=path, group=group)

    def _add_to_queue(self, rec):
        """Add record to queue."""
        # find insertion index for record (queue sorted in ascending order)
        # we expect new records to go near the end (most recent)
        queue = self.queues[rec.group]
        with self._record_lock:
            for k, (kkey, kpath) in enumerate(reversed(queue)):
                if rec.key > kkey:
                    # we've found the insertion point at index k from end
                    break
                elif rec.path == kpath:
                    # already in ringbuffer, so simply return
                    return
            else:
                # new key is oldest (or queue is empty),
                # needs to be put at beginning
                queue.appendleft((rec.key, rec.path))
                k = None
            if k is not None:
                # insert record at index k
                queue.rotate(k)
                queue.append((rec.key, rec.path))
                queue.rotate(-k)

    def _remove_from_queue(self, rec):
        """Remove record from queue."""
        queue = self.queues[rec.group]
        with self._record_lock:
            queue.remove((rec.key, rec.path))

    def _expire_oldest_from_group(self, group):
        """Expire oldest record from group and delete corresponding file."""
        # oldest file is at start of sorted records deque
        # (don't just popleft on the queue because we want to call
        #  _remove_from_queue, which is overridden by subclasses)
        with self._record_lock:
            key, path = self.queues[group][0]
            rec = self.records.pop(path)
            self._remove_from_queue(rec)

        if self.verbose:
            print('Expired {0}'.format(rec.path))

        # delete file
        if not self.dryrun:
            try:
                os.remove(rec.path)
            except OSError:
                # path doesn't exist like we thought it did, oh well
                if self.verbose:
                    traceback.print_exc()
        # try to clean up directory in case it is empty
        head, tail = os.path.split(rec.path)
        try:
            os.rmdir(head)
        except OSError:
            # directory not empty, just move on
            pass

    def _expire(self, group):
        """Expire records until ringbuffer constraint is met."""
        # must override with mixins for any expiration to occur
        pass

    def _add_record(self, rec):
        """Add a record to the ringbuffer and expire old ones if necesssary."""
        with self._record_lock:
            # make sure record does not already exist, remove if it does
            if rec.path in self.records:
                if self.verbose:
                    msg = (
                        'Adding record for {0} but it already exists in'
                        ' ringbuffer, modify instead.'
                    ).format(rec.path)
                    print(msg)
                self._modify_record(rec)
            # add record to dict so record information can be looked up by path
            self.records[rec.path] = rec
            # add record to expiration queue
            self._add_to_queue(rec)

            if self.verbose:
                print('Added {0}'.format(rec.path))
            # expire oldest files until size constraint is met
            self._expire(rec.group)

    def _modify_record(self, rec):
        """Modify a record in the ringbuffer, return whether it was done."""
        with self._record_lock:
            if rec.path not in self.records:
                # don't have record in ringbuffer when we should, add instead
                if self.verbose:
                    msg = (
                        'Missing modified file {0} from ringbuffer, adding'
                        ' instead.'
                    ).format(rec.path)
                    print(msg)
                self._add_record(rec)
                return True
            # nothing to do otherwise
            return False

    def _remove_record(self, path):
        """Remove a record from the ringbuffer."""
        # get and remove record id if path is in the ringbuffer, return if not
        with self._record_lock:
            try:
                rec = self.records.pop(path)
            except KeyError:
                # we probably got here from a FileDeletedEvent after expiring
                # an old record, but in any case, no harm to just ignore
                return
            # remove record from ringbuffer
            self._remove_from_queue(rec)

        if self.verbose:
            print('Removed {0}'.format(rec.path))

    def add_files(self, paths, sort=True):
        """Create file records from paths and add to ringbuffer."""
        # get records and add from oldest to newest by key (time)
        records = (self._get_file_record(p) for p in paths)
        # filter out invalid paths (can't extract a time, doesn't exist)
        records = (r for r in records if r is not None)
        if sort:
            records = sorted(records)
        for rec in records:
            self._add_record(rec)

    def modify_files(self, paths, sort=True):
        """Create file records from paths and update in ringbuffer."""
        # get records and add from oldest to newest by key (time)
        records = (self._get_file_record(p) for p in paths)
        # filter out invalid paths (can't extract a time, doesn't exist)
        records = (r for r in records if r is not None)
        if sort:
            records = sorted(records)
        for rec in records:
            self._modify_record(rec)

    def remove_files(self, paths):
        """Retrieve file records from paths and remove from ringbuffer."""
        for p in paths:
            self._remove_record(p)

    def on_created(self, event):
        """Add new file to ringbuffer."""
        self.add_files([event.src_path])

    def on_deleted(self, event):
        """Remove file from ringbuffer if it was deleted externally."""
        self.remove_files([event.src_path])

    def on_modified(self, event):
        """Update modified file in ringbuffer."""
        self.modify_files([event.src_path])

    def on_moved(self, event):
        """Track moved file in ringbuffer."""
        self.remove_files([event.src_path])
        self.add_files([event.dest_path])


class CountExpirer(object):
    """Ringbuffer handler mixin to track the number of files in each channel.

    When the count threshold of a channel is exceeded, the oldest files in that
    channel are deleted until the count constraint is met.

    """

    def __init__(self, *args, **kwargs):
        """Create a ringbuffer handler."""
        self.count = kwargs.pop('count')
        super(CountExpirer, self).__init__(*args, **kwargs)

    def status(self):
        """Return status string about state of the ringbuffer."""
        status = super(CountExpirer, self).status()
        try:
            max_count = max(len(q) for q in self.queues.values())
        except ValueError:
            max_count = 0
        pct_full = int(float(max_count)/self.count*100)
        return ', '.join((status, '{0}% count'.format(pct_full)))

    def _expire(self, group):
        """Expire records until file count constraint is met."""
        with self._record_lock:
            queue = self.queues[group]
            while (len(queue) > self.count):
                self._expire_oldest_from_group(group)
        super(CountExpirer, self)._expire(group)


class SizeExpirer(object):
    """Ringbuffer handler mixin to track the space used by all channels.

    This expirer tracks the amount of space that new or modified files consume
    for all channels together. When the space threshold is exceeded, the oldest
    file of any channel is deleted (unless it would empty the channel) until
    the size constraint is met.

    """

    def __init__(self, *args, **kwargs):
        """Create a ringbuffer handler."""
        self.size = kwargs.pop('size')
        self.active_size = 0
        super(SizeExpirer, self).__init__(*args, **kwargs)

    def status(self):
        """Return status string about state of the ringbuffer."""
        status = super(SizeExpirer, self).status()
        pct_full = int(float(self.active_size)/self.size*100)
        return ', '.join((status, '{0}% size'.format(pct_full)))

    def _add_to_queue(self, rec):
        """Add record to queue, tracking file size."""
        with self._record_lock:
            super(SizeExpirer, self)._add_to_queue(rec)
            self.active_size += rec.size

    def _remove_from_queue(self, rec):
        """Remove record from queue, tracking file size."""
        with self._record_lock:
            super(SizeExpirer, self)._remove_from_queue(rec)
            self.active_size -= rec.size

    def _expire_oldest(self, group):
        """Expire oldest record overall, preferring group if tied."""
        with self._record_lock:
            # remove oldest regardless of group unless it would empty group,
            # but prefer `group` if tie
            removal_group = group
            # oldest file is at start of sorted records deque
            try:
                oldest_key, oldest_path = self.queues[group][0]
            except IndexError:
                oldest_key = float('inf')
            for grp in self.queues.keys():
                if grp != group:
                    queue = self.queues[grp]
                    if len(queue) > 1:
                        key, path = queue[0]
                        if key < oldest_key:
                            oldest_key = key
                            removal_group = grp
            self._expire_oldest_from_group(removal_group)

    def _expire(self, group):
        """Expire records until overall file size constraint is met."""
        with self._record_lock:
            while (self.active_size > self.size):
                self._expire_oldest(group)
        super(SizeExpirer, self)._expire(group)

    def _modify_record(self, rec):
        """Modify a record in the ringbuffer."""
        with self._record_lock:
            # have parent handle cases where we actually need to add or delete
            handled = super(SizeExpirer, self)._modify_record(rec)
            if not handled:
                # if we're here, we know that the record exists and needs to
                # be modified (in this case, update the size)
                oldrec = self.records[rec.path]

                self.records[rec.path] = rec
                self.active_size -= oldrec.size
                self.active_size += rec.size

                if self.verbose:
                    print('Updated {0}'.format(rec.path))


class TimeExpirer(object):
    """Ringbuffer handler mixin to track the time span of each channel.

    This handler tracks the sample timestamp of files in each channel. When the
    duration threshold of a channel is exceeded (newest timestamp minus
    oldest), the oldest files in the channel are deleted until the duration
    constraint is met.

    """

    def __init__(self, *args, **kwargs):
        """Create a ringbuffer handler."""
        # duration is time span in milliseconds
        self.duration = kwargs.pop('duration')
        super(TimeExpirer, self).__init__(*args, **kwargs)

    @staticmethod
    def _queue_duration(queue):
        """Get time span in milliseconds of files in a queue."""
        try:
            oldkey, _ = queue[0]
            newkey, _ = queue[-1]
        except IndexError:
            return 0
        return (newkey - oldkey)

    def status(self):
        """Return status string about state of the ringbuffer."""
        status = super(TimeExpirer, self).status()
        try:
            max_duration = max(
                self._queue_duration(q) for q in self.queues.values()
            )
        except ValueError:
            max_duration = 0
        pct_full = int(float(max_duration)/self.duration*100)
        return ', '.join((status, '{0}% duration'.format(pct_full)))

    def _expire(self, group):
        """Expire records until time span constraint is met."""
        with self._record_lock:
            queue = self.queues[group]
            while (self._queue_duration(queue) > self.duration):
                self._expire_oldest_from_group(group)
        super(TimeExpirer, self)._expire(group)


def DigitalRFRingbufferHandler(size=None, count=None, duration=None, **kwargs):
    """Create ringbuffer handler given constraints.

    Parameters
    ----------

    size : float | int | None
        Size of the ringbuffer in bytes. Negative values are used to
        indicate all available space except the given amount. If None, no
        size constraint is used.

    count : int | None
        Maximum number of files *for each channel*. If None, no count
        constraint is used.

    duration : int | float | None
        Maximum time span *for each channel* in milliseconds. If None, no
        duration constraint is used.


    Other Parameters
    ----------------

    verbose : bool
        If True, print debugging info about the files that are created and
        deleted and how much space they consume.

    dryrun : bool
        If True, do not actually delete files when expiring them from the
        ringbuffer. Use for testing only!

    starttime : datetime.datetime
        Data covering this time or after will be included. This has no
        effect on property files.

    endtime : datetime.datetime
        Data covering this time or earlier will be included. This has no
        effect on property files.

    include_drf : bool
        If True, include Digital RF files. If False, ignore Digital RF
        files.

    include_dmd : bool
        If True, include Digital Metadata files. If False, ignore Digital
        Metadata files.

    """
    if size is None and count is None and duration is None:
        errstr = 'One of `size`, `count`, or `duration` must not be None.'
        raise ValueError(errstr)

    bases = (DigitalRFRingbufferHandlerBase, )
    # add mixins in this particular order for expected results
    if size is not None:
        bases = (SizeExpirer, ) + bases
        kwargs['size'] = size
    if duration is not None:
        bases = (TimeExpirer, ) + bases
        kwargs['duration'] = duration
    if count is not None:
        bases = (CountExpirer, ) + bases
        kwargs['count'] = count

    # now create the class with the desired mixins
    docstring = (
        """Event handler for implementing a ringbuffer of Digital RF files.

        This class inherits from a base class (DigitalRFRingbufferHandlerBase)
        and some expirer mixins determined from the class factor arguments.
        The expirers determine when a file needs to be expired from the
        ringbuffer based on size, count, or duration constraints.

        """
    )
    cls = type('DigitalRFRingbufferHandler', bases, {'__doc__': docstring})

    return cls(**kwargs)


class DigitalRFRingbuffer(object):
    """Monitor a directory and delete old Digital RF files when space is full.

    This class combines an event handler and a file system observer. It
    monitors a directory and its subdirectories for new Digital RF and Digital
    Metadata files. When the ringbuffer threshold in size, count, or duration
    is exceeded, the oldest files are deleted until the constraint is met.

    """

    def __init__(
        self, path, size=-200e6, count=None, duration=None,
        verbose=False, status_interval=10, dryrun=False,
        starttime=None, endtime=None, include_drf=True, include_dmd=True,
    ):
        """Create Digital RF ringbuffer object. Use start/run method to begin.

        Parameters
        ----------

        path : str
            Directory in which the ringbuffer is enforced.

        size : float | int | None
            Size of the ringbuffer in bytes. Negative values are used to
            indicate all available space except the given amount. If None, no
            size constraint is used.

        count : int | None
            Maximum number of files *for each channel*. If None, no count
            constraint is used.

        duration : int | float | None
            Maximum time span *for each channel* in milliseconds. If None, no
            duration constraint is used.


        Other Parameters
        ----------------

        verbose : bool
            If True, print debugging info about the files that are created and
            deleted and how much space they consume.

        status_interval : None | int
            Interval in seconds between printing of status updates. If None,
            do not print status updates.

        dryrun : bool
            If True, do not actually delete files when expiring them from the
            ringbuffer. Use for testing only!

        starttime : datetime.datetime
            Data covering this time or after will be included. This has no
            effect on property files.

        endtime : datetime.datetime
            Data covering this time or earlier will be included. This has no
            effect on property files.

        include_drf : bool
            If True, include Digital RF files. If False, ignore Digital RF
            files.

        include_dmd : bool
            If True, include Digital Metadata files. If False, ignore Digital
            Metadata files.

        """
        self.path = os.path.abspath(path)
        self.size = size
        self.count = count
        self.duration = duration
        self.verbose = verbose
        self.status_interval = status_interval
        self.dryrun = dryrun
        self.starttime = starttime
        self.endtime = endtime
        self.include_drf = include_drf
        self.include_dmd = include_dmd
        self._start_time = None
        self._task_threads = []

        if self.size is None and self.count is None and self.duration is None:
            errstr = 'One of `size`, `count`, or `duration` must not be None.'
            raise ValueError(errstr)

        if not self.include_drf and not self.include_dmd:
            errstr = 'One of `include_drf` or `include_dmd` must be True.'
            raise ValueError(errstr)

        if self.status_interval is None:
            self.status_interval = float('inf')

        if self.size is not None:
            if self.size < 0:
                # get available space and reduce it by the (negative) size
                # value to get the actual size to use
                root = self.path
                while not os.path.isdir(root):
                    root = os.path.dirname(root)
                statvfs = os.statvfs(root)
                bytes_available = statvfs.f_frsize*statvfs.f_bavail
                if os.path.isdir(self.path):
                    existing = ilsdrf(
                        self.path, starttime=self.starttime,
                        endtime=self.endtime,
                        include_drf=self.include_drf,
                        include_dmd=self.include_dmd,
                        include_drf_properties=False,
                        include_dmd_properties=False,
                    )
                    bytes_available += sum(
                        os.stat(p).st_size for p in existing
                    )
                self.size = max(bytes_available + self.size, 0)

        self.event_handler = DigitalRFRingbufferHandler(
            size=self.size, count=self.count, duration=self.duration,
            verbose=self.verbose, dryrun=self.dryrun,
            starttime=self.starttime, endtime=self.endtime,
            include_drf=self.include_drf, include_dmd=self.include_dmd,
        )

        self._init_observer()

    def _init_observer(self):
        self.observer = DirWatcher(self.path)
        self.observer.schedule(self.event_handler, self.path, recursive=True)

    def _add_existing_files(self):
        """Add existing files on disk to ringbuffer."""
        # pause dispatching while we add existing files so files are added
        # to the ringbuffer in the correct order
        with self.observer.paused_dispatching():
            # add existing files to ringbuffer handler
            existing = ilsdrf(
                self.path, starttime=self.starttime, endtime=self.endtime,
                include_drf=self.include_drf, include_dmd=self.include_dmd,
                include_drf_properties=False, include_dmd_properties=False,
            )
            # do not sort because existing will already be sorted and we
            # don't want to convert to a list
            self.event_handler.add_files(existing, sort=False)

    def start(self):
        """Start ringbuffer process."""
        self._start_time = datetime.datetime.utcnow().replace(microsecond=0)

        print('{0} | Starting observer.'.format(self._start_time))
        sys.stdout.flush()

        # start observer to add new files
        self.observer.start()

        # add files that already existed before the observer started
        # (do it in another thread so we can get to join())
        thread = threading.Thread(target=self._add_existing_files)
        thread.daemon = True
        thread.start()
        self._task_threads.append(thread)

    def _verify_ringbuffer_files(self, inbuffer):
        """Verify ringbuffer's `inbuffer` set of files with files on disk."""
        # get set of all files that should be in the ringbuffer right away
        # so we duplicate as few files from new events as possible
        # events that happen while we build this file set can be duplicated
        # when we verify the ringbuffer state below, but that's ok
        ondisk = set(ilsdrf(
            self.path, starttime=self.starttime, endtime=self.endtime,
            include_drf=self.include_drf, include_dmd=self.include_dmd,
            include_drf_properties=False, include_dmd_properties=False,
        ))

        # now any file in inbuffer that is not in ondisk is a missed or
        # duplicate deletion event, so remove those files
        deletions = inbuffer - ondisk
        self.event_handler.remove_files(deletions)

        # any file in ondisk that is not in inbuffer is a missed or duplicate
        # creation event, so add those files
        creations = ondisk - deletions
        self.event_handler.add_files(creations, sort=True)

        # any file in both ondisk and inbuffer could have a missed modify
        # event, so trigger a modify event for those files
        possibly_modified = inbuffer & ondisk
        self.event_handler.modify_files(possibly_modified, sort=True)

    def _restart(self):
        """Restart observer using existing event handlers."""
        # get list of files currently in ringbuffer before we modify it
        # so we can detect missed events from after crash until ondisk
        # file list is complete
        inbuffer = set(self.event_handler.records.keys())

        # make a new observer and start it ASAP
        self._init_observer()
        self.observer.start()

        # verify existing state of ringbuffer
        # (do it in another thread so we can get to join())
        thread = threading.Thread(
            target=self._verify_ringbuffer_files,
            kwargs=dict(inbuffer=inbuffer),
        )
        thread.daemon = True
        thread.start()
        self._task_threads.append(thread)

    def join(self):
        """Wait until a KeyboardInterrupt is received to stop ringbuffer."""
        try:
            while True:
                now = datetime.datetime.utcnow().replace(microsecond=0)
                interval = int((now - self._start_time).total_seconds())
                if (interval % self.status_interval) == 0:
                    status = self.event_handler.status()
                    print('{0} | ({1})'.format(now, status))
                    sys.stdout.flush()

                if not self.observer.all_alive():
                    # if not all threads of the observer are alive,
                    # reinitialize and restart
                    print(
                        'Found stopped thread, reinitializing and restarting.'
                    )
                    sys.stdout.flush()
                    # first make sure all task threads have stopped
                    for thread in self._task_threads:
                        while thread.is_alive():
                            print('Waiting for task thread to finish.')
                            sys.stdout.flush()
                            thread.join(1)
                    del self._task_threads[:]
                    self._restart()

                time.sleep(1)
        except (KeyboardInterrupt, SystemExit):
            self.stop()
            sys.stdout.write('\n')
            sys.stdout.flush()
        self.observer.join()

    def run(self):
        """Start ringbuffer and wait for a KeyboardInterrupt to stop."""
        self.start()
        self.join()

    def stop(self):
        """Stop ringbuffer process."""
        self.observer.stop()

    def __str__(self):
        """String describing ringbuffer."""
        amounts = []
        if self.size is not None:
            amounts.append('{0} bytes'.format(self.size))
        if self.count is not None:
            amounts.append('{0} files'.format(self.count))
        if self.duration is not None:
            amounts.append('{0} s'.format(self.duration/1e3))
        s = 'DigitalRFRingbuffer of ({0}) in {1}'.format(
            ', '.join(amounts), self.path,
        )
        return s


def _build_ringbuffer_parser(Parser, *args):
    desc = (
        'Enforce ringbuffer of Digital RF and Digital Metadata files. When'
        ' the space threshold is exceeded, the oldest files are deleted until'
        ' the size constraint is met.'
    )
    parser = Parser(*args, description=desc)

    parser.add_argument(
        'path',
        help='Directory in which to enforce ringbuffer.',
    )
    parser.add_argument(
        '-z', '--size', default=None,
        help='''Size of ringbuffer, in bytes or using unit symbols (e.g 100GB).
                Negative values are used to indicate all available space except
                the given amount. (default: -200MB if no count or duration)''',
    )
    parser.add_argument(
        '-c', '--count', type=int, default=None,
        help='''Max file count for each channel. (default: %(default)s)''',
    )
    parser.add_argument(
        '-l', '--duration', default=None,
        help='''Max duration for each channel in seconds.
                (default: %(default)s)''',
    )
    parser.add_argument(
        '-v', '--verbose', action='store_true',
        help='Print the name new/deleted files and the space consumed.',
    )
    parser.add_argument(
        '-p', '--status_interval', type=int, default=10,
        help='''Interval in seconds between printing of status updates.
                (default: %(default)s)''',
    )
    parser.add_argument(
        '-n', '--dryrun', action='store_true',
        help='Do not delete files when expiring them from the ringbuffer.',
    )

    parser = list_drf._add_time_group(parser)

    includegroup = parser.add_argument_group(title='include')
    includegroup.add_argument(
        '--nodrf', dest='include_drf', action='store_false',
        help='''Do not ringbuffer Digital RF HDF5 files.
                (default: False)''',
    )
    includegroup.add_argument(
        '--nodmd', dest='include_dmd', action='store_false',
        help='''Do not ringbuffer Digital Metadata HDF5 files.
                (default: False)''',
    )

    parser.set_defaults(func=_run_ringbuffer)

    return parser


def _run_ringbuffer(args):
    # parse size string into number of bytes
    if args.size == '':
        args.size = None
    if args.size is not None:
        suffixes = OrderedDict([
            ('B', 1),
            ('KB', 1000**1),
            ('KiB', 1024**1),
            ('MB', 1000**2),
            ('MiB', 1024**2),
            ('GB', 1000**3),
            ('GiB', 1024**3),
            ('TB', 1000**4),
            ('TiB', 1024**4),
            ('PB', 1000**5),
            ('PiB', 1024**5),
        ])
        m = re.match('(?P<num>\-?\d+\.?\d*)(?P<suf>\D*)', args.size)
        if not m:
            raise ValueError('Size string not recognized. '
                             'Use number followed by suffix.')
        sizenum = eval(m.group('num'))
        suf = m.group('suf').strip()
        if not suf:
            args.size = sizenum
        elif suf in suffixes:
            args.size = sizenum*suffixes[suf]
        else:
            raise ValueError('Size suffix not recognized. Use one of:\n'
                             '{0}'.format(suffixes.keys()))
    elif args.count is None and args.duration is None:
        args.size = -200e6

    # evaluate duration to float, from seconds to milliseconds
    if args.duration is not None:
        args.duration = float(eval(args.duration))*1e3

    if args.starttime is not None:
        args.starttime = util.parse_identifier_to_time(args.starttime)
    if args.endtime is not None:
        args.endtime = util.parse_identifier_to_time(
            args.endtime, ref_datetime=args.starttime,
        )

    kwargs = vars(args).copy()
    del kwargs['func']
    ringbuffer = DigitalRFRingbuffer(**kwargs)
    if args.dryrun:
        print('DRY RUN (files will not be deleted):')
    print(ringbuffer)
    print('Type Ctrl-C to quit.')
    sys.stdout.flush()
    ringbuffer.run()


if __name__ == "__main__":
    from argparse import ArgumentParser

    parser = _build_ringbuffer_parser(ArgumentParser)
    args = parser.parse_args()
    args.func(args)



#!python
# ----------------------------------------------------------------------------
# Copyright (c) 2017 Massachusetts Institute of Technology (MIT)
# All rights reserved.
#
# Distributed under the terms of the BSD 3-clause license.
#
# The full license is in the LICENSE file, distributed with this software.
# ----------------------------------------------------------------------------
"""Script for analyzing pseudorandom-coded waveforms.

See the following paper for a description and application of the technique:

Vierinen, J., Chau, J. L., Pfeffer, N., Clahsen, M., and Stober, G.,
Coded continuous wave meteor radar, Atmos. Meas. Tech., 9, 829-839,
doi:10.5194/amt-9-829-2016, 2016.

"""
import datetime
import glob
import itertools
import math
import os
import time
from argparse import ArgumentParser

import numpy as np
import scipy.signal

import digital_rf as drf


def create_pseudo_random_code(clen=10000, seed=0):
    """
    seed is a way of reproducing the random code without
    having to store all actual codes. the seed can then
    act as a sort of station_id.

    """
    np.random.seed(seed)
    phases = np.array(
        np.exp(1.0j * 2.0 * math.pi * np.random.random(clen)),
        dtype=np.complex64,
    )
    return(phases)


def periodic_convolution_matrix(envelope, rmin=0, rmax=100):
    """
    we imply that the number of measurements is equal to the number of elements
    in code

    """
    L = len(envelope)
    ridx = np.arange(rmin, rmax)
    A = np.zeros([L, rmax-rmin], dtype=np.complex64)
    for i in np.arange(L):
        A[i, :] = envelope[(i-ridx) % L]
    result = {}
    result['A'] = A
    result['ridx'] = ridx
    return(result)


B_cache = 0
r_cache = 0
B_cached = False
def create_estimation_matrix(code, rmin=0, rmax=1000, cache=True):
    global B_cache
    global r_cache
    global B_cached

    if not cache or not B_cached:
        r_cache = periodic_convolution_matrix(
            envelope=code, rmin=rmin, rmax=rmax,
        )
        A = r_cache['A']
        Ah = np.transpose(np.conjugate(A))
        B_cache = np.dot(np.linalg.inv(np.dot(Ah, A)), Ah)
        r_cache['B'] = B_cache
        B_cached = True
        return(r_cache)
    else:
        return(r_cache)


def analyze_prc(
    dirn='', channel='hfrx', idx0=0, an_len=1000000, clen=10000, station=0,
    Nranges=1000, rfi_rem=True, cache=True,
):
    r"""Analyze pseudorandom code transmission for a block of data.

    idx0 = start idx
    an_len = analysis length
    clen = code length
    station = random seed for pseudorandom code
    cache = Do we cache (\conj(A^T)\*A)^{-1}\conj{A}^T for linear least squares
        solution (significant speedup)
    rfi_rem = Remove RFI (whiten noise).

    """
    if type(dirn) is str:
        g = drf.DigitalRFReader(dirn)
    else:
        g = dirn

    code = create_pseudo_random_code(clen=clen, seed=station)
    N = an_len / clen
    res = np.zeros([N, Nranges], dtype=np.complex64)
    r = create_estimation_matrix(code=code, cache=cache, rmax=Nranges)
    B = r['B']
    spec = np.zeros([N, Nranges], dtype=np.complex64)

    for i in np.arange(N):
        z = g.read_vector_c81d(idx0 + i * clen, clen, channel)
        z = z - np.median(z)  # remove dc
        res[i, :] = np.dot(B, z)
    for i in np.arange(Nranges):
        spec[:, i] = np.fft.fftshift(np.fft.fft(
            scipy.signal.blackmanharris(N) * res[:, i]
        ))

    if rfi_rem:
        median_spec = np.zeros(N, dtype=np.float32)
        for i in np.arange(N):
            median_spec[i] = np.median(np.abs(spec[i, :]))
        for i in np.arange(Nranges):
            spec[:, i] = spec[:, i] / median_spec[:]
    ret = {}
    ret['res'] = res
    ret['spec'] = spec
    return(ret)


if __name__ == '__main__':
    import matplotlib
    matplotlib.use('Agg')
    import matplotlib.pyplot as plt

    desc = """Script for analyzing pseudorandom-coded waveforms.

    See the following paper for a description and application of the technique:

    Vierinen, J., Chau, J. L., Pfeffer, N., Clahsen, M., and Stober, G.,
    Coded continuous wave meteor radar, Atmos. Meas. Tech., 9, 829-839,
    doi:10.5194/amt-9-829-2016, 2016.

    """

    parser = ArgumentParser(description=desc)

    parser.add_argument(
        'datadir', help='''Data directory to analyze.''',
    )
    parser.add_argument(
        '-c', '--ch', default='hfrx',
        help='''Channel name of data to analyze. (default: %(default)s)'''
    )
    parser.add_argument(
        '-o', '--out', dest='outdir', default='{datadir}/prc_analysis',
        help='''Processed output directory. (default: %(default)s)''',
    )
    parser.add_argument(
        '-x', '--delete_old', action='store_true', default=False,
        help='''Delete existing processed files.''',
    )
    parser.add_argument(
        '-n', '--analysis_length', dest='anlen', type=int, default=6000000,
        help='''Analysis length. (default: %(default)s)''',
    )
    parser.add_argument(
        '-l', '--code_length', dest='codelen', type=int, default=10000,
        help='''Code length. (default: %(default)s)''',
    )
    parser.add_argument(
        '-s', '--station', type=int, default=0,
        help='''Station ID for code (seed). (default: %(default)s)''',
    )
    parser.add_argument(
        '-r', '--nranges', type=int, default=1000,
        help='''Number of range gates. (default: %(default)s)''',
    )

    op = parser.parse_args()

    op.datadir = os.path.abspath(op.datadir)
    # join outdir to datadir to allow for relative path, normalize
    op.outdir = os.path.abspath(op.outdir.format(datadir=op.datadir))
    if not os.path.isdir(op.outdir):
        os.makedirs(op.outdir)
    datpath = os.path.join(op.outdir, 'last.dat')
    if op.delete_old:
        for f in itertools.chain(
            glob.iglob(datpath),
            glob.iglob(os.path.join(op.outdir, '*.png')),
        ):
            os.remove(f)

    d = drf.DigitalRFReader(op.datadir)
    sr = d.get_properties(op.ch)['samples_per_second']
    b = d.get_bounds(op.ch)
    idx = np.array(b[0])
    if os.path.isfile(datpath):
        fidx = np.fromfile(datpath, dtype=np.int)
        if b[0] <= fidx:
            idx = fidx

    while True:
        if idx + op.anlen > b[1]:
            print('waiting for more data, sleeping.')
            time.sleep(op.anlen / sr)
            b = d.get_bounds(op.ch)
            continue

        try:
            res = analyze_prc(
                d, channel=op.ch, idx0=idx, an_len=op.anlen, clen=op.codelen,
                station=op.station, Nranges=op.nranges,
                cache=True, rfi_rem=False,
            )
            plt.clf()

            M = 10.0 * np.log10((np.abs(res['spec'])))
            plt.pcolormesh(np.transpose(M), vmin=(np.median(M) - 1.0))

            plt.colorbar()
            plt.title(
                datetime.datetime.utcfromtimestamp(idx / sr).strftime(
                    '%Y-%m-%d %H:%M:%S'
                )
            )
            plt.savefig(os.path.join(
                op.outdir, 'spec-{0:06d}.png'.format(int(np.uint64(idx / sr))),
            ))
            print('%d' % (idx))
        except IOError:
            print('IOError, skipping.')
        idx = idx + op.anlen
        idx.tofile(datpath)



#!python
# ----------------------------------------------------------------------------
# Copyright (c) 2017 Massachusetts Institute of Technology (MIT)
# All rights reserved.
#
# Distributed under the terms of the BSD 3-clause license.
#
# The full license is in the LICENSE file, distributed with this software.
# ----------------------------------------------------------------------------
"""Convert Digital RF 1 formatted data to the current version.

Use verify_digital_rf_upconvert.py if you want to test the conversion.

$Id$
"""

# standard python imports
import argparse
import glob
import math
import os
import warnings

import h5py
import numpy as np

# Millstone imports
import digital_rf
from digital_rf import digital_rf_deprecated_hdf5  # for reading old formatter

read_len = 1000000  # default read len


if __name__ == '__main__':
    # command line interface
    parser = argparse.ArgumentParser(
        description='''Convert Digital RF 1 formatted data to the current
                       version.'''
    )
    parser.add_argument(
        '--source', metavar='sourceDir', required=True,
        help='''Source top level directory containing Digital RF channels to be
                converted''',
    )
    parser.add_argument(
        '--target', metavar='targetDir', required=True,
        help='''Target top level directory where Digital RF channels will be
                written after upconversion''',
    )
    parser.add_argument(
        '--dir_secs', metavar='Subdir_cadence_seconds', type=int, default=3600,
        help='''Number of seconds of data to store per subdirectory.
                Default is 3600 (one hour)''',
    )
    parser.add_argument(
        '--file_millisecs', metavar='File_cadence_milliseconds', type=int,
        default=1000,
        help='''Number of milliseconds of data to store per file.
                Default is 1000 (one second)''',
    )
    parser.add_argument(
        '--gzip', metavar='GZIP compression 0-9.', type=int, default=0,
        help='''Level of GZIP compression 1-9.  Default=0 (no compression)''',
    )
    parser.add_argument(
        '--checksum', action='store_true',
        help='''Set this flag to turn on Hdf5 checksums''',
    )
    args = parser.parse_args()

    reader = digital_rf_deprecated_hdf5.read_hdf5(args.source)

    # convert each channel separately
    for channel in reader.get_channels():
        print('working on channel %s' % (channel))
        metaDict = reader.get_rf_file_metadata(channel)
        bounds = reader.get_bounds(channel)

        # this code only works if the sample rate is an integer
        sample_rate = metaDict['sample_rate'][0]
        if math.fmod(sample_rate, 1.0) != 0.0:
            errstr = (
                'Cannot guess the numerator and denominator with a fractional'
                ' sample rate %f'
            )
            raise ValueError(errstr % sample_rate)
        sample_rate_numerator = int(sample_rate)
        sample_rate_denominator = int(1)

        # read critical metadata
        is_complex = bool(metaDict['is_complex'][0])
        num_subchannels = metaDict['num_subchannels'][0]
        uuid_str = str(metaDict['uuid_str'])

        # get first sample to find dtype
        data = reader.read_vector_raw(bounds[0], 1, channel)
        this_dtype = data.dtype
        cont_blocks = reader.get_continuous_blocks(
            bounds[0], bounds[1], channel)
        subdir = os.path.join(args.target, channel)
        if not os.access(subdir, os.R_OK):
            os.makedirs(subdir)

        # extract sampler_util metadata in any metadata* files
        metadataFiles = glob.glob(os.path.join(
            args.source, channel, 'metadata*.h5',
        ))
        metadataFiles.sort()
        if metadataFiles:
            # create metadata dir, dmd object, and write channel metadata
            mddir = os.path.join(subdir, 'metadata')
            if not os.path.exists(mddir):
                os.makedirs(mddir)
            mdo = digital_rf.DigitalMetadataWriter(
                metadata_dir=mddir,
                subdir_cadence_secs=args.dir_secs,
                file_cadence_secs=1,
                sample_rate_numerator=sample_rate_numerator,
                sample_rate_denominator=sample_rate_denominator,
                file_name='metadata',
            )
            for filepath in metadataFiles:
                try:
                    with h5py.File(filepath, 'r') as mdfile:
                        md = dict(
                            uuid_str=uuid_str,
                            sample_rate_numerator=sample_rate_numerator,
                            sample_rate_denominator=sample_rate_denominator,
                            # put in a list because we want the data to be a
                            # 1-D array and it would be a single value o/w
                            center_frequencies=[
                                mdfile['center_frequencies'][()].reshape((-1,))
                            ],
                        )
                        try:
                            # try for newer added metadata, usrp_id and friends
                            receiver_dict = dict(
                                description='UHD USRP source using GNU Radio',
                                gain=str(mdfile['usrp_gain'][()]),
                                id=str(mdfile['usrp_id'][()]),
                                stream_args=str(
                                    mdfile['usrp_stream_args'][()]
                                ),
                                subdev=str(mdfile['usrp_subdev'][()]),
                            )
                        except KeyError:
                            # fallback to original usrp metadata
                            receiver_dict = dict(
                                description='UHD USRP source using GNU Radio',
                                id=str(mdfile['usrp_ip'][()]),
                            )
                        md['receiver'] = receiver_dict
                except (IOError, KeyError):
                    # file is bad, warn and ignore
                    errstr = 'Skipping bad metadata file {0}.'
                    warnings.warn(errstr.format(filepath))
                else:
                    fname = os.path.basename(filepath)
                    t = np.longdouble(fname.split('@', 1)[1][:-3])
                    s = int((t * sample_rate_numerator)
                            / sample_rate_denominator)
                    mdo.write(samples=s, data=md)

        # create a drf 2 writer
        writer = digital_rf.DigitalRFWriter(
            subdir, this_dtype, args.dir_secs, args.file_millisecs, bounds[0],
            sample_rate_numerator, sample_rate_denominator,
            uuid_str=uuid_str, is_complex=is_complex,
            num_subchannels=num_subchannels, compression_level=args.gzip,
            checksum=args.checksum, marching_periods=False,
        )

        # write all the data
        for startSample, sampleLen in cont_blocks:
            thisSample = startSample
            endSample = (startSample + sampleLen) - 1
            while thisSample < endSample:
                this_len = min((endSample - thisSample) + 1, read_len)
                data = reader.read_vector_raw(
                    thisSample, this_len, channel)
                writer.rf_write(data, thisSample - bounds[0])
                thisSample += this_len

        writer.close()



#!python
# ----------------------------------------------------------------------------
# Copyright (c) 2017 Massachusetts Institute of Technology (MIT)
# All rights reserved.
#
# Distributed under the terms of the BSD 3-clause license.
#
# The full license is in the LICENSE file, distributed with this software.
# ----------------------------------------------------------------------------
"""Record data from synchronized USRPs in Digital RF format."""
from __future__ import print_function

import math
import os
import re
import sys
import time
from argparse import (
    Action, ArgumentParser, Namespace, RawDescriptionHelpFormatter,
)
from ast import literal_eval
from datetime import datetime, timedelta
from fractions import Fraction
from itertools import chain, cycle, islice, repeat
from subprocess import call
from textwrap import TextWrapper, dedent, fill

import numpy as np
import pytz
from gnuradio import filter, gr, uhd
from gnuradio.filter import firdes

import digital_rf as drf
import gr_digital_rf as gr_drf


def evalint(s):
    """Evaluate string to an integer."""
    return int(eval(s, {}, {}))


def evalfloat(s):
    """Evaluate string to a float."""
    return float(eval(s, {}, {}))


def noneorstr(s):
    """Turn empty or 'none' string to None."""
    if s.lower() in ('', 'none'):
        return None
    else:
        return s


def noneorbool(s):
    """Turn empty or 'none' string to None, all others to boolean."""
    if s.lower() in ('', 'none'):
        return None
    elif s.lower() in ('true', 't', 'yes', 'y', '1'):
        return True
    else:
        return False


class Extend(Action):
    """Action to split comma-separated arguments and add to a list."""

    def __init__(self, option_strings, dest, type=None, **kwargs):
        if type is not None:
            itemtype = type
        else:
            def itemtype(s):
                return s

        def split_string_and_cast(s):
            return [itemtype(a.strip()) for a in s.strip().split(',')]

        super(Extend, self).__init__(
            option_strings, dest, type=split_string_and_cast, **kwargs
        )

    def __call__(self, parser, namespace, values, option_string=None):
        cur_list = getattr(namespace, self.dest, [])
        if cur_list is None:
            cur_list = []
        cur_list.extend(values)
        setattr(namespace, self.dest, cur_list)


class Thor(object):
    """Record data from synchronized USRPs in DigitalRF format."""

    def __init__(
        self, datadir, mboards=[], subdevs=['A:A'],
        chs=['ch0'], centerfreqs=[100e6], lo_offsets=[0],
        lo_sources=[''], lo_exports=[None],
        gains=[0], bandwidths=[0], antennas=[''],
        samplerate=1e6, dec=1,
        dev_args=['recv_buff_size=100000000', 'num_recv_frames=512'],
        stream_args=[], tune_args=[],
        sync=True, sync_source='external',
        stop_on_dropped=False, realtime=False,
        file_cadence_ms=1000, subdir_cadence_s=3600, metadata={}, uuid=None,
        verbose=True, test_settings=True,
    ):
        options = locals()
        del options['self']
        op = self._parse_options(**options)
        self.op = op

        # test usrp device settings, release device when done
        if op.test_settings:
            if op.verbose:
                print('Initialization: testing device settings.')
            u = self._usrp_setup()
            del u

    @staticmethod
    def _parse_options(**kwargs):
        """Put all keyword options in a namespace and normalize them."""
        op = Namespace(**kwargs)

        op.nmboards = len(op.mboards) if len(op.mboards) > 0 else 1
        op.nchs = len(op.chs)
        # repeat arguments as necessary
        op.subdevs = list(islice(cycle(op.subdevs), 0, op.nmboards))
        op.centerfreqs = list(islice(cycle(op.centerfreqs), 0, op.nchs))
        op.lo_offsets = list(islice(cycle(op.lo_offsets), 0, op.nchs))
        op.lo_sources = list(islice(cycle(op.lo_sources), 0, op.nchs))
        op.lo_exports = list(islice(cycle(op.lo_exports), 0, op.nchs))
        op.gains = list(islice(cycle(op.gains), 0, op.nchs))
        op.bandwidths = list(islice(cycle(op.bandwidths), 0, op.nchs))
        op.antennas = list(islice(cycle(op.antennas), 0, op.nchs))

        # create device_addr string to identify the requested device(s)
        op.mboard_strs = []
        for n, mb in enumerate(op.mboards):
            if re.match(r'[^0-9]+=.+', mb):
                idtype, mb = mb.split('=')
            elif re.match(
                r'[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}', mb
            ):
                idtype = 'addr'
            elif (
                re.match(r'usrp[123]', mb) or re.match(r'b2[01]0', mb)
                or re.match(r'x3[01]0', mb)
            ):
                idtype = 'type'
            elif re.match(r'[0-9A-Fa-f]{1,}', mb):
                idtype = 'serial'
            else:
                idtype = 'name'
            if len(op.mboards) == 1:
                # do not use identifier numbering if only using one mainboard
                s = '{type}={mb}'.format(type=idtype, mb=mb.strip())
            else:
                s = '{type}{n}={mb}'.format(type=idtype, n=n, mb=mb.strip())
            op.mboard_strs.append(s)

        if op.verbose:
            opstr = dedent('''\
                Main boards: {mboard_strs}
                Subdevices: {subdevs}
                Channel names: {chs}
                Frequency: {centerfreqs}
                LO frequency offset: {lo_offsets}
                LO source: {lo_sources}
                LO export: {lo_exports}
                Gain: {gains}
                Bandwidth: {bandwidths}
                Antenna: {antennas}
                Device arguments: {dev_args}
                Stream arguments: {stream_args}
                Tune arguments: {tune_args}
                Sample rate: {samplerate}
                Data dir: {datadir}
                Metadata: {metadata}
                UUID: {uuid}
            ''').strip().format(**op.__dict__)
            print(opstr)

        # check that subdevice specifications are unique per-mainboard
        for sd in op.subdevs:
            sds = sd.split()
            if len(set(sds)) != len(sds):
                errstr = (
                    'Invalid subdevice specification: "{0}". '
                    'Each subdevice specification for a given mainboard must '
                    'be unique.'
                )
                raise ValueError(errstr.format(sd))

        # sanity check: # of total subdevs should be same as # of channels
        op.mboards_bychan = []
        op.subdevs_bychan = []
        op.mboardnum_bychan = []
        mboards = op.mboards if op.mboards else ['default']
        for mbnum, (mb, sd) in enumerate(zip(mboards, op.subdevs)):
            sds = sd.split()
            mbs = list(repeat(mb, len(sds)))
            mbnums = list(repeat(mbnum, len(sds)))
            op.mboards_bychan.extend(mbs)
            op.subdevs_bychan.extend(sds)
            op.mboardnum_bychan.extend(mbnums)
        if len(op.subdevs_bychan) != op.nchs:
            raise ValueError(
                'Number of device channels does not match the number of '
                'channel names provided.'
            )

        return op

    def _usrp_setup(self):
        """Create, set up, and return USRP source object."""
        op = self.op
        # create usrp source block
        if op.dec > 1:
            op.cpu_format = 'fc32'
        else:
            op.cpu_format = 'sc16'
        op.otw_format = 'sc16'
        u = uhd.usrp_source(
            device_addr=','.join(chain(op.mboard_strs, op.dev_args)),
            stream_args=uhd.stream_args(
                cpu_format=op.cpu_format,
                otw_format=op.otw_format,
                channels=range(len(op.chs)),
                args=','.join(op.stream_args)
            ),
        )

        # set clock and time source if synced
        if op.sync:
            try:
                u.set_clock_source(op.sync_source, uhd.ALL_MBOARDS)
                u.set_time_source(op.sync_source, uhd.ALL_MBOARDS)
            except RuntimeError:
                errstr = (
                    "Unknown sync_source option: '{0}'. Must be one of {1}."
                ).format(op.sync_source, u.get_clock_sources(0))
                raise ValueError(errstr)

        # set mainboard options
        for mb_num in range(op.nmboards):
            u.set_subdev_spec(op.subdevs[mb_num], mb_num)
        # set global options
        # sample rate
        u.set_samp_rate(float(op.samplerate))
        # read back actual value
        samplerate = u.get_samp_rate()
        # calculate longdouble precision sample rate
        # (integer division of clock rate)
        cr = u.get_clock_rate()
        srdec = int(round(cr / samplerate))
        samplerate_ld = np.longdouble(cr) / srdec
        op.samplerate = samplerate_ld
        sr_rat = Fraction(cr).limit_denominator() / srdec
        op.samplerate_num = sr_rat.numerator
        op.samplerate_den = sr_rat.denominator

        # set per-channel options
        # set command time so settings are synced
        COMMAND_DELAY = 0.2
        cmd_time = u.get_time_now() + uhd.time_spec(COMMAND_DELAY)
        u.set_command_time(cmd_time, uhd.ALL_MBOARDS)
        for ch_num in range(op.nchs):
            # local oscillator sharing settings
            lo_source = op.lo_sources[ch_num]
            if lo_source:
                try:
                    u.set_lo_source(lo_source, uhd.ALL_LOS, ch_num)
                except RuntimeError:
                    errstr = (
                        "Unknown LO source option: '{0}'. Must be one of {1},"
                        " or it may not be possible to set the LO source on"
                        " this daughterboard."
                    ).format(lo_source, u.get_lo_sources(uhd.ALL_LOS, ch_num))
                    raise ValueError(errstr)
            lo_export = op.lo_exports[ch_num]
            if lo_export is not None:
                if not lo_source:
                    errstr = (
                        'Channel {0}: must set an LO source in order to set'
                        ' LO export.'
                    ).format(ch_num)
                    raise ValueError(errstr)
                u.set_lo_export_enabled(lo_export, uhd.ALL_LOS, ch_num)
            # center frequency and tuning offset
            tune_res = u.set_center_freq(
                uhd.tune_request(
                    op.centerfreqs[ch_num], op.lo_offsets[ch_num],
                    args=uhd.device_addr(','.join(op.tune_args)),
                ),
                ch_num,
            )
            # store actual values from tune result
            op.centerfreqs[ch_num] = (
                tune_res.actual_rf_freq - tune_res.actual_dsp_freq
            )
            op.lo_offsets[ch_num] = tune_res.actual_dsp_freq
            # gain
            u.set_gain(op.gains[ch_num], ch_num)
            # bandwidth
            bw = op.bandwidths[ch_num]
            if bw:
                u.set_bandwidth(bw, ch_num)
            # antenna
            ant = op.antennas[ch_num]
            if ant:
                try:
                    u.set_antenna(ant, ch_num)
                except RuntimeError:
                    errstr = (
                        "Unknown RX antenna option: '{0}'. Must be one of {1}."
                    ).format(ant, u.get_antennas(ch_num))
                    raise ValueError(errstr)

        # commands are done, clear time
        u.clear_command_time(uhd.ALL_MBOARDS)
        time.sleep(COMMAND_DELAY)

        # read back actual channel settings
        for ch_num in range(op.nchs):
            if op.lo_sources[ch_num]:
                op.lo_sources[ch_num] = u.get_lo_source(uhd.ALL_LOS, ch_num)
            if op.lo_exports[ch_num] is not None:
                op.lo_exports[ch_num] = u.get_lo_export_enabled(
                    uhd.ALL_LOS, ch_num,
                )
            op.gains[ch_num] = u.get_gain(ch_num)
            op.bandwidths[ch_num] = u.get_bandwidth(chan=ch_num)
            op.antennas[ch_num] = u.get_antenna(chan=ch_num)

        if op.verbose:
            print('Using the following devices:')
            chinfostrs = [
                'Motherboard: {mb_id} ({mb_addr}) | Daughterboard: {db_name}',
                'Subdev: {sub} | Antenna: {ant} | Gain: {gain} | Rate: {sr}',
                'Frequency: {freq:.3f} ({lo_off:+.3f}) | Bandwidth: {bw}',
            ]
            if any(op.lo_sources) or any(op.lo_exports):
                chinfostrs.append(
                    'LO source: {lo_source} | LO export: {lo_export}'
                )
            chinfo = '\n'.join(['  ' + l for l in chinfostrs])
            for ch_num in range(op.nchs):
                header = '---- {0} '.format(op.chs[ch_num])
                header += '-' * (78 - len(header))
                print(header)
                usrpinfo = dict(u.get_usrp_info(chan=ch_num))
                info = {}
                info['mb_id'] = usrpinfo['mboard_id']
                mba = op.mboards_bychan[ch_num]
                if mba == 'default':
                    mba = usrpinfo['mboard_serial']
                info['mb_addr'] = mba
                info['db_name'] = usrpinfo['rx_subdev_name']
                info['sub'] = op.subdevs_bychan[ch_num]
                info['ant'] = op.antennas[ch_num]
                info['bw'] = op.bandwidths[ch_num]
                info['freq'] = op.centerfreqs[ch_num]
                info['gain'] = op.gains[ch_num]
                info['lo_off'] = op.lo_offsets[ch_num]
                info['lo_source'] = op.lo_sources[ch_num]
                info['lo_export'] = op.lo_exports[ch_num]
                info['sr'] = op.samplerate
                print(chinfo.format(**info))
                print('-' * 78)

        return u

    def run(self, starttime=None, endtime=None, duration=None, period=10):
        op = self.op

        # window in seconds that we allow for setup time so that we don't
        # issue a start command that's in the past when the flowgraph starts
        SETUP_TIME = 10

        # print current time and NTP status
        if op.verbose:
            call(('timedatectl', 'status'))

        # parse time arguments
        st = drf.util.parse_identifier_to_time(starttime)
        if st is not None:
            # find next suitable start time by cycle repeat period
            now = datetime.utcnow()
            now = now.replace(tzinfo=pytz.utc)
            soon = now + timedelta(seconds=SETUP_TIME)
            diff = max(soon - st, timedelta(0)).total_seconds()
            periods_until_next = (diff - 1) // period + 1
            st = st + timedelta(seconds=periods_until_next * period)

            if op.verbose:
                ststr = st.strftime('%a %b %d %H:%M:%S %Y')
                stts = (st - drf.util.epoch).total_seconds()
                print('Start time: {0} ({1})'.format(ststr, stts))

        et = drf.util.parse_identifier_to_time(endtime, ref_datetime=st)
        if et is not None:
            if op.verbose:
                etstr = et.strftime('%a %b %d %H:%M:%S %Y')
                etts = (et - drf.util.epoch).total_seconds()
                print('End time: {0} ({1})'.format(etstr, etts))

            if ((et < (pytz.utc.localize(datetime.utcnow())
                       + timedelta(seconds=SETUP_TIME)))
               or (st is not None and et <= st)):
                raise ValueError('End time is before launch time!')

        if op.realtime:
            r = gr.enable_realtime_scheduling()

            if op.verbose:
                if r == gr.RT_OK:
                    print('Realtime scheduling enabled')
                else:
                    print('Note: failed to enable realtime scheduling')

        # create data directory so ringbuffer code can be started while waiting
        # to launch
        if not os.path.isdir(op.datadir):
            os.makedirs(op.datadir)

        # wait for the start time if it is not past
        while (st is not None) and (
            (st - pytz.utc.localize(datetime.utcnow())) >
                timedelta(seconds=SETUP_TIME)
        ):
            ttl = int((
                st - pytz.utc.localize(datetime.utcnow())
            ).total_seconds())
            if (ttl % 10) == 0:
                print('Standby {0} s remaining...'.format(ttl))
                sys.stdout.flush()
            time.sleep(1)

        # get UHD USRP source
        u = self._usrp_setup()

        # after USRP setup, get output settings that depend on decimation rate
        samplerate_out = op.samplerate / op.dec
        samplerate_num_out = op.samplerate_num
        samplerate_den_out = op.samplerate_den * op.dec
        if op.dec > 1:
            sample_dtype = '<c8'

            taps = firdes.low_pass_2(
                1.0, float(op.samplerate), float(samplerate_out / 2.0),
                float(0.2 * samplerate_out), 80.0,
                window=firdes.WIN_BLACKMAN_hARRIS
            )
        else:
            sample_dtype = np.dtype([('r', '<i2'), ('i', '<i2')])

        # force creation of the RX streamer ahead of time with a start/stop
        # (after setting time/clock sources, before setting the
        # device time)
        # this fixes timing with the B210
        u.start()
        # need to wait >0.1 s (constant in usrp_source_impl.c) for start/stop
        # to actually take effect, so sleep a bit
        time.sleep(0.2)
        u.stop()
        time.sleep(0.2)

        # set device time
        tt = time.time()
        if op.sync:
            # wait until time 0.2 to 0.5 past full second, then latch
            # we have to trust NTP to be 0.2 s accurate
            while tt - math.floor(tt) < 0.2 or tt - math.floor(tt) > 0.3:
                time.sleep(0.01)
                tt = time.time()
            if op.verbose:
                print('Latching at ' + str(tt))
            # waits for the next pps to happen
            # (at time math.ceil(tt))
            # then sets the time for the subsequent pps
            # (at time math.ceil(tt) + 1.0)
            u.set_time_unknown_pps(uhd.time_spec(math.ceil(tt) + 1.0))
        else:
            u.set_time_now(uhd.time_spec(tt))

        # set launch time
        # (at least 1 second out so USRP time is set, time to set up flowgraph)
        if st is not None:
            lt = st
        else:
            now = pytz.utc.localize(datetime.utcnow())
            lt = now.replace(microsecond=0) + timedelta(seconds=2)
        ltts = (lt - drf.util.epoch).total_seconds()
        # adjust launch time forward so it falls on an exact sample since epoch
        lt_samples = np.ceil(ltts * samplerate_out)
        ltts = lt_samples / samplerate_out
        lt = drf.util.sample_to_datetime(lt_samples, samplerate_out)
        if op.verbose:
            ltstr = lt.strftime('%a %b %d %H:%M:%S.%f %Y')
            print('Launch time: {0} ({1})'.format(ltstr, repr(ltts)))
        # command launch time
        ct_td = lt - drf.util.epoch
        ct_secs = ct_td.total_seconds() // 1.0
        ct_frac = ct_td.microseconds / 1000000.0
        u.set_start_time(
            uhd.time_spec(ct_secs) + uhd.time_spec(ct_frac)
        )

        # populate flowgraph one channel at a time
        fg = gr.top_block()
        for k in range(op.nchs):
            mbnum = op.mboardnum_bychan[k]
            # create digital RF sink
            dst = gr_drf.digital_rf_channel_sink(
                channel_dir=os.path.join(op.datadir, op.chs[k]),
                dtype=sample_dtype,
                subdir_cadence_secs=op.subdir_cadence_s,
                file_cadence_millisecs=op.file_cadence_ms,
                sample_rate_numerator=samplerate_num_out,
                sample_rate_denominator=samplerate_den_out,
                start=lt_samples,
                ignore_tags=False,
                is_complex=True,
                num_subchannels=1,
                uuid_str=op.uuid,
                center_frequencies=op.centerfreqs[k],
                metadata=dict(
                    # receiver metadata for USRP
                    receiver=dict(
                        description='UHD USRP source using GNU Radio',
                        info=dict(u.get_usrp_info(chan=k)),
                        antenna=op.antennas[k],
                        bandwidth=op.bandwidths[k],
                        center_freq=op.centerfreqs[k],
                        clock_rate=u.get_clock_rate(mboard=mbnum),
                        clock_source=u.get_clock_source(mboard=mbnum),
                        gain=op.gains[k],
                        id=op.mboards_bychan[k],
                        lo_offset=op.lo_offsets[k],
                        otw_format=op.otw_format,
                        samp_rate=u.get_samp_rate(),
                        stream_args=','.join(op.stream_args),
                        subdev=op.subdevs_bychan[k],
                        time_source=u.get_time_source(mboard=mbnum),
                    ),
                ),
                is_continuous=True,
                compression_level=0,
                checksum=False,
                marching_periods=True,
                stop_on_skipped=op.stop_on_dropped,
                debug=op.verbose,
            )

            if op.dec > 1:
                # create low-pass filter
                lpf = filter.freq_xlating_fir_filter_ccf(
                    op.dec, taps, 0.0, float(op.samplerate)
                )

                # connections for usrp->lpf->drf
                connections = ((u, k), (lpf, 0), (dst, 0))
            else:
                # connections for usrp->drf
                connections = ((u, k), (dst, 0))

            # make channel connections in flowgraph
            fg.connect(*connections)

        # start the flowgraph once we are near the launch time
        # (start too soon and device buffers might not yet be flushed)
        # (start too late and device might not be able to start in time)
        while ((lt - pytz.utc.localize(datetime.utcnow()))
                > timedelta(seconds=1.2)):
            time.sleep(0.1)
        fg.start()

        # wait until end time or until flowgraph stops
        if et is None and duration is not None:
            et = lt + timedelta(seconds=duration)
        try:
            if et is None:
                fg.wait()
            else:
                # sleep until end time nears
                while(pytz.utc.localize(datetime.utcnow()) <
                        et - timedelta(seconds=2)):
                    time.sleep(1)
                else:
                    # issue stream stop command at end time
                    ct_td = et - drf.util.epoch
                    ct_secs = ct_td.total_seconds() // 1.0
                    ct_frac = ct_td.microseconds / 1000000.0
                    u.set_command_time(
                        (uhd.time_spec(ct_secs) + uhd.time_spec(ct_frac)),
                        uhd.ALL_MBOARDS,
                    )
                    stop_enum = uhd.stream_cmd.STREAM_MODE_STOP_CONTINUOUS
                    u.issue_stream_cmd(uhd.stream_cmd(stop_enum))
                    u.clear_command_time(uhd.ALL_MBOARDS)
                    # sleep until after end time
                    time.sleep(2)
        except KeyboardInterrupt:
            # catch keyboard interrupt and simply exit
            pass
        fg.stop()
        # need to wait for the flowgraph to clean up, otherwise it won't exit
        fg.wait()
        print('done')
        sys.stdout.flush()


if __name__ == '__main__':
    scriptname = os.path.basename(sys.argv[0])

    formatter = RawDescriptionHelpFormatter(scriptname)
    width = formatter._width

    title = 'THOR (The Haystack Observatory Recorder)'
    copyright = 'Copyright (c) 2017 Massachusetts Institute of Technology'
    shortdesc = 'Record data from synchronized USRPs in DigitalRF format.'
    desc = '\n'.join((
        '*'*width,
        '*{0:^{1}}*'.format(title, width-2),
        '*{0:^{1}}*'.format(copyright, width-2),
        '*{0:^{1}}*'.format('', width-2),
        '*{0:^{1}}*'.format(shortdesc, width-2),
        '*'*width,
    ))

    usage = (
        '%(prog)s [-m MBOARD] [-d SUBDEV] [-c CH] [-y ANT] [-f FREQ]'
        ' [-F OFFSET] \\\n'
        '{0:8}[-g GAIN] [-b BANDWIDTH] [-r RATE] [options] DIR\n'.format('')
    )

    epi_pars = [
        '''\
        Arguments in the "mainboard" and "channel" groups accept multiple
        values, allowing multiple mainboards and channels to be specified.
        Multiple arguments can be provided by repeating the argument flag, by
        passing a comma-separated list of values, or both. Within each argument
        group, parameters will be grouped in the order in which they are given
        to form the complete set of parameters for each mainboard/channel. For
        any argument with fewer values given than the number of
        mainboards/channels, its values will be extended by repeatedly cycling
        through the values given up to the needed number.
        ''',
        '''\
        Arguments in other groups apply to all mainboards/channels (including
        the sample rate and decimation factor).
        ''',
        '''\
        Example usage:
        ''',
    ]
    epi_pars = [fill(dedent(s), width) for s in epi_pars]

    egtw = TextWrapper(
        width=(width - 2), break_long_words=False, break_on_hyphens=False,
        subsequent_indent=' ' * (len(scriptname) + 1),
    )
    egs = [
        '''\
        {0} -m 192.168.20.2 -d "A:A A:B" -c h,v -f 95e6 -r 100e6/24
        /data/test
        ''',
        '''\
        {0} -m 192.168.10.2 -d "A:0" -c ch1 -y "TX/RX" -f 20e6 -F 10e3 -g 20
        -b 0 -r 1e6 /data/test
        ''',
    ]
    egs = [' \\\n'.join(egtw.wrap(dedent(s.format(scriptname)))) for s in egs]
    epi = '\n' + '\n\n'.join(epi_pars + egs) + '\n'

    # parse options
    parser = ArgumentParser(description=desc, usage=usage, epilog=epi,
                            formatter_class=RawDescriptionHelpFormatter)

    parser.add_argument(
        '--version', action='version',
        version='THOR 3.1, using digital_rf {0}'.format(drf.__version__),
    )
    parser.add_argument(
        '-q', '--quiet', dest='verbose', action='store_false',
        help='''Reduce text output to the screen. (default: False)''',
    )
    parser.add_argument(
        '--notest', dest='test_settings', action='store_false',
        help='''Do not test USRP settings until experiment start.
                (default: False)''',
    )

    dirgroup = parser.add_mutually_exclusive_group(required=True)
    dirgroup.add_argument(
        'datadir', nargs='?', default=None,
        help='''Data directory, to be filled with channel subdirectories.''',
    )
    dirgroup.add_argument(
        '-o', '--out', dest='outdir', default=None,
        help='''Data directory, to be filled with channel subdirectories.''',
    )

    mbgroup = parser.add_argument_group(title='mainboard')
    mbgroup.add_argument(
        '-m', '--mainboard', dest='mboards', action=Extend,
        help='''Mainboard address. (default: first device found)''',
    )
    mbgroup.add_argument(
        '-d', '--subdevice', dest='subdevs', action=Extend,
        help='''USRP subdevice string. (default: "A:A")''',
    )

    chgroup = parser.add_argument_group(title='channel')
    chgroup.add_argument(
        '-c', '--channel', dest='chs', action=Extend,
        help='''Channel names to use in data directory. (default: "ch0")''',
    )
    chgroup.add_argument(
        '-f', '--centerfreq', dest='centerfreqs', action=Extend,
        type=evalfloat,
        help='''Center frequency in Hz. (default: 100e6)''',
    )
    chgroup.add_argument(
        '-F', '--lo_offset', dest='lo_offsets', action=Extend, type=evalfloat,
        help='''Frontend tuner offset from center frequency, in Hz.
                (default: 0)''',
    )
    chgroup.add_argument(
        '--lo_source', dest='lo_sources', action=Extend, type=noneorstr,
        help='''Local oscillator source. Typically 'None'/'' (do not set),
                'internal' (e.g. LO1 for CH1, LO2 for CH2),
                'companion' (e.g. LO2 for CH1, LO1 for CH2), or
                'external' (neighboring board via connector).
                (default: '')''',
    )
    chgroup.add_argument(
        '--lo_export', dest='lo_exports', action=Extend, type=noneorbool,
        help='''Whether to export the LO's source to the external connector.
                Can be 'None'/'' to skip the channel, otherwise it can be
                'True' or 'False' provided the LO source is set.
                (default: None)''',
    )
    chgroup.add_argument(
        '-g', '--gain', dest='gains', action=Extend, type=evalfloat,
        help='''Gain in dB. (default: 0)''',
    )
    chgroup.add_argument(
        '-b', '--bandwidth', dest='bandwidths', action=Extend, type=evalfloat,
        help='''Frontend bandwidth in Hz. (default: 0 == frontend default)''',
    )
    chgroup.add_argument(
        '-y', '--antenna', dest='antennas', action=Extend, type=noneorstr,
        help='''Name of antenna to select on the frontend.
                (default: frontend default))''',
    )

    recgroup = parser.add_argument_group(title='receiver')
    recgroup.add_argument(
        '-r', '--samplerate', dest='samplerate', type=evalfloat,
        help='''Sample rate in Hz. (default: 1e6)''',
    )
    recgroup.add_argument(
        '-i', '--dec', dest='dec', type=int,
        help='''Integrate and decimate by this factor.
                (default: 1)''',
    )
    recgroup.add_argument(
        '-A', '--devargs', dest='dev_args', action=Extend,
        help='''Device arguments, e.g. "master_clock_rate=30e6".
                (default: 'recv_buff_size=100000000,num_recv_frames=512')''',
    )
    recgroup.add_argument(
        '-a', '--streamargs', dest='stream_args', action=Extend,
        help='''Stream arguments, e.g. "peak=0.125,fullscale=1.0".
                (default: '')''',
    )
    recgroup.add_argument(
        '-T', '--tuneargs', dest='tune_args', action=Extend,
        help='''Tune request arguments, e.g. "mode_n=integer,int_n_step=100e3".
                (default: '')''',
    )
    recgroup.add_argument(
        '--sync_source', dest='sync_source',
        help='''Clock and time source for all mainboards.
                (default: external)''',
    )
    recgroup.add_argument(
        '--nosync', dest='sync', action='store_false',
        help='''No syncing with external clock. (default: False)''',
    )
    recgroup.add_argument(
        '--stop_on_dropped', dest='stop_on_dropped', action='store_true',
        help='''Stop on dropped packet. (default: %(default)s)''',
    )
    recgroup.add_argument(
        '--realtime', dest='realtime', action='store_true',
        help='''Enable realtime scheduling if possible.
                (default: %(default)s)''',
    )

    timegroup = parser.add_argument_group(title='time')
    timegroup.add_argument(
        '-s', '--starttime', dest='starttime',
        help='''Start time of the experiment as datetime (if in ISO8601 format:
                2016-01-01T15:24:00Z) or Unix time (if float/int).
                (default: start ASAP)''',
    )
    timegroup.add_argument(
        '-e', '--endtime', dest='endtime',
        help='''End time of the experiment as datetime (if in ISO8601 format:
                2016-01-01T16:24:00Z) or Unix time (if float/int).
                (default: wait for Ctrl-C)''',
    )
    timegroup.add_argument(
        '-l', '--duration', dest='duration', type=evalint,
        help='''Duration of experiment in seconds. When endtime is not given,
                end this long after start time. (default: wait for Ctrl-C)''',
    )
    timegroup.add_argument(
        '-p', '--cycle-length', dest='period', type=evalint,
        help='''Repeat time of experiment cycle. Align to start of next cycle
                if start time has passed. (default: 10)''',
    )

    drfgroup = parser.add_argument_group(title='digital_rf')
    drfgroup.add_argument(
        '-n', '--file_cadence_ms', dest='file_cadence_ms', type=evalint,
        help='''Number of milliseconds of data per file.
                (default: 1000)''',
    )
    drfgroup.add_argument(
        '-N', '--subdir_cadence_s', dest='subdir_cadence_s', type=evalint,
        help='''Number of seconds of data per subdirectory.
                (default: 3600)''',
    )
    drfgroup.add_argument(
        '--metadata', action=Extend, metavar='{KEY}={VALUE}',
        help='''Key, value metadata pairs to include with data.
                (default: "")''',
    )
    drfgroup.add_argument(
        '--uuid', dest='uuid',
        help='''Unique ID string for this data collection.
                (default: random)''',
    )

    op = parser.parse_args()

    if op.datadir is None:
        op.datadir = op.outdir
    del op.outdir

    # remove redundant arguments in dev_args, stream_args, tune_args
    if op.dev_args is not None:
        try:
            dev_args_dict = dict([a.split('=') for a in op.dev_args])
        except ValueError:
            raise ValueError(
                'Device arguments must be {KEY}={VALUE} pairs.'
            )
        op.dev_args = [
            '{0}={1}'.format(k, v) for k, v in dev_args_dict.iteritems()
        ]
    if op.stream_args is not None:
        try:
            stream_args_dict = dict([a.split('=') for a in op.stream_args])
        except ValueError:
            raise ValueError(
                'Stream arguments must be {KEY}={VALUE} pairs.'
            )
        op.stream_args = [
            '{0}={1}'.format(k, v) for k, v in stream_args_dict.iteritems()
        ]
    if op.tune_args is not None:
        try:
            tune_args_dict = dict([a.split('=') for a in op.tune_args])
        except ValueError:
            raise ValueError(
                'Tune request arguments must be {KEY}={VALUE} pairs.'
            )
        op.tune_args = [
            '{0}={1}'.format(k, v) for k, v in tune_args_dict.iteritems()
        ]

    # convert metadata strings to a dictionary
    if op.metadata is not None:
        metadata_dict = {}
        for a in op.metadata:
            try:
                k, v = a.split('=')
            except ValueError:
                k = None
                v = a
            try:
                v = literal_eval(v)
            except ValueError:
                pass
            if k is None:
                metadata_dict.setdefault('metadata', []).append(v)
            else:
                metadata_dict[k] = v
        op.metadata = metadata_dict

    # ignore test_settings option if no starttime is set (starting right now)
    if op.starttime is None:
        op.test_settings = False

    options = {k: v for k, v in op._get_kwargs() if v is not None}
    runopts = {
        k: options.pop(k) for k in list(options.keys())
        if k in ('starttime', 'endtime', 'duration', 'period')
    }
    thor = Thor(**options)
    thor.run(**runopts)
